/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-html.c: HTML display routines
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_STATVFS
#include <sys/statvfs.h>
#endif
#include <sys/vfs.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gtkhtml/gtkhtml-embedded.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gtkhtml/gtkhtml-embedded.h>
#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlengine-search.h>

#include <gal/util/e-unicode-i18n.h>

#include "gui.h"
#include "gui-transfer.h"
#include "gui-prefs.h"
#include "gui-install.h"
#include "mirror.h"
#include "subscription.h"

#include "libredcarpet.h"
#include "rc-debug-misc.h"

#include "rc-package-spec.h"

#include "gui-html.h"
#include "gui-util.h"
#include "gui-uri.h"

/* for search_box_activated */
#include "gui-glade.h"

#define  LEFT_CORNER_IMG "<img border=no width=16 height=36 src=\"file://"SHAREDIR"/left-corner.png\">"
#define RIGHT_CORNER_IMG "<img border=no width=16 height=36 src=\"file://"SHAREDIR"/right-corner.png\">"
#define  LEFT_CORNER_RED "<img border=no width=16 height=36 src=\"file://"SHAREDIR"/left-corner-red.png\">"
#define RIGHT_CORNER_RED "<img border=no width=16 height=36 src=\"file://"SHAREDIR"/right-corner-red.png\">"

typedef struct _GUIHtmlButtonCallbackStruct GUIHtmlButtonCallbackStruct;

struct _GUIHtmlButtonCallbackStruct {
    GtkHTML *html;
    const char *button_title;
    const char *uri;
    GUIChannel *channel;
    gboolean init_sensitive;
};

static GUIHtmlButtonCallbackStruct *rc_gui_make_button_callback_struct(
    GtkHTML *html,
    const char *button_title,
    const char *uri,
    GUIChannel *channel,
    gboolean init_sensitive);

static void rc_gui_handle_uri_button_cb (GtkWidget *button,
                                         gpointer data);


#define RENDER_PACKAGE_FLAGS_NONE      0
#define RENDER_PACKAGE_FLAGS_OLDER     (1 << 0)
#define RENDER_PACKAGE_FLAGS_DISCOLOR  (1 << 1)

#undef BROKEN_OLDER

static gchar *
rc_gui_render_yellow_bar (const gchar *color, const gchar *font_color,
                          const gchar *left_corner, const gchar *icon,
                          const gchar *text1, const gchar *text2,
                          const gchar *right_corner)
{
    gchar *html;

    html = g_strdup_printf (
        "<table width=100%% cellpadding=0 cellspacing=0>"
        "<tr bgcolor=%s>"
        "<td width=0 bgcolor=#ffffff>%s</td>"
        "%s%s%s%s%s"
        "<td width=100%%><font size=+1 color=%s><b>%s</b></font></td>"
        "%s<font color=%s>%s</font>%s"
        "<td width=0 bgcolor=#ffffff>%s</td>"
        "</tr>"
        "</table>",
        color,
        left_corner, icon ? "<td width=0>" : "",
        icon ? "<img width=32 height=32 src=" : "",
        icon ? icon : "", icon ? ">" : "",
        icon ? "</td>" : "", font_color, text1,
        text2 ? "<td width=0 nowrap>" : "",
        font_color,
        text2 ? text2 : "",
        text2 ? "</td>" : "",
        right_corner);

    return (html);
} /* rc_gui_render_yellow_bar */

static void
append_info(RCResolverInfo *info, gpointer user_data)
{
    char **html_info = user_data;
    char *new_html_info;
    char *msg = rc_resolver_info_to_str(info);

    if (!(*html_info))
        new_html_info = g_strdup(msg);
    else
        new_html_info = g_strconcat(*html_info, "<br>", msg, NULL);

    g_free(*html_info);
    g_free(msg);

    *html_info = new_html_info;
}

void
rc_gui_render_dependency_failure(GUIChannel *current_channel,
                                 RCResolver *resolver)
{
    GtkHTML *html;
    GtkHTMLStream *stream;
    char *html_data;
    char *tmp;

    /* Do not allow select all/none or subscribe/unsubscribe here */
    rc_gui_sensitize_select_items(SELECT_NEITHER);
    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);
    rc_gui_set_menubar_sensitive (FALSE);
    
    
    html = GTK_HTML(
        gtk_object_get_user_data(GTK_OBJECT(GET_WIDGET("verify_head"))));
    stream = gtk_html_begin(html);
    html_data = g_strdup_printf (
        "<html><body bgcolor=\"#000000\" marginwidth=0 marginheight=1>"
        "<table cellspacing=0 cellpadding=0 width=\"100%%\" border=0>"
        "<tr>"
        "<td align=left><img src=\"file://"SHAREDIR"/download.png\">"
        "</td>"
        "<td align=left width=\"100%%\">"
        "<font size=\"+2\" color=\"#ffffff\">"
        "<b>%s</b></font>"      /* "Failed Dependencies" */
        "</td></tr></table>"
        "</body></html>",
        U_("<!--html-->Failed Dependencies")); /* i18n */
    gtk_html_write(html, stream, html_data, strlen(html_data));
    g_free (html_data);

    gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

    html = GTK_HTML(
        gtk_object_get_user_data(GTK_OBJECT(GET_WIDGET("verify_html"))));

    stream = gtk_html_begin (html);

    html_data = g_strdup_printf(
        "<html>"
        "<body bgcolor=#ffffff text=#000000 "
        "      marginwidth=0 marginheight=0>"
        "<p>%s</p>",
        U_("<!--html-->Red Carpet was unable to successfully locate all the "
           "necessary files to complete this request. The errors and "
           "conflicts that occured are listed here, along with a suggested "
           "action to take."));
    gtk_html_write(html, stream, html_data, strlen(html_data));
    g_free(html_data);


    tmp = g_strdup_printf(
        "<a href=\"updater:send_bug_report?Dependencies failed\">"
        "<font size=-1 color=#ffffff>%s</a>",
        U_("Click here to send a dependency report"));

    html_data = rc_gui_render_yellow_bar(
        "#ff0000", "#ffffff", LEFT_CORNER_RED,
        "file://"SHAREDIR"/warning.png",
        U_("Failed Dependencies"), tmp, RIGHT_CORNER_RED);

    g_free(tmp);
    gtk_html_write (html, stream, html_data, strlen (html_data));
    g_free(html_data);

    {
        RCResolverQueue *queue;

        /* FIXME: Choose a best invalid queue */
        queue = (RCResolverQueue *) resolver->invalid_queues->data;

        html_data = NULL;
        rc_resolver_context_foreach_info(
            queue->context, NULL, -1, append_info, &html_data);
        gtk_html_write(html, stream, html_data, strlen(html_data));
        g_free(html_data);
    }
    
#if 0
    if (rr->unresolved_reqs) {
        GSList *l;

        l = rr->unresolved_reqs;
        while (l) {
            RCDepRequire *req = (RCDepRequire *) l->data;
            GSList *deplines, *iter;

            iter = deplines = rc_dep_require_to_user_strings(
                appinfo->system_packages,
                req,
                TRUE);
            while (iter) {
                html_data = g_strdup_printf(
                    "<li>%s",
                    (gchar *) iter->data);
                gtk_html_write (html, stream, html_data, strlen (html_data));
                g_free (html_data);

                iter = iter->next;
            }

            g_slist_foreach (deplines, (GFunc) g_free, NULL);
            g_slist_free (deplines);


            l = l->next;
        }
    }

    if (rr->packages_conflicting) {
        GSList *l;

        l = rr->packages_conflicting;
        while (l) {
            RCDepConflict *dc = (RCDepConflict *) l->data;
            gchar *s;

            s = rc_dep_conflict_to_user_string (dc);

            html_data = g_strdup_printf(
                "<li>%s",
                s);
            g_free (s);

            gtk_html_write (html, stream, html_data, strlen (html_data));
            
            l = l->next;
        }
    }
#endif

    html_data = "</body></html>";
    gtk_html_write (html, stream, html_data, strlen (html_data));
    
    gtk_html_end (html, stream, GTK_HTML_STREAM_OK);
} /* rc_gui_render_dependency_failure */

/* ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** */

static char *
info_to_html (RCResolverInfo *info)
{
    char *str = NULL, *pkgs;
    
    g_return_val_if_fail (info != NULL, NULL);

    switch (info->type) {
        
    case RC_RESOLVER_INFO_TYPE_NEEDED_BY:
        pkgs = rc_resolver_info_packages_to_str (info, TRUE);
        str = g_strdup_printf (U_("needed by: <b>%s</b>"), pkgs);
        g_free (pkgs);
        break;
        
    case RC_RESOLVER_INFO_TYPE_CONFLICTS_WITH:
        pkgs = rc_resolver_info_packages_to_str (info, TRUE);
        str = g_strdup_printf (U_("conflicts with: <b>%s</b>"), pkgs);
        g_free (pkgs);
        break;
        
    case RC_RESOLVER_INFO_TYPE_DEPENDS_ON:
        pkgs = rc_resolver_info_packages_to_str (info, TRUE);
        str = g_strdup_printf (U_("depends on: <b>%s</b>"), pkgs);
        g_free (pkgs);
        break;

    default:
        str = rc_resolver_info_to_str (info);

    }

    return str;
}

struct RenderInfoData {
    GtkHTML *html;
    GtkHTMLStream *stream;
    const char *template;
};

static void
render_info_cb (RCResolverInfo *info, gpointer user_data)
{
    struct RenderInfoData *data = user_data;
    char *info_str, *total_str;

    info_str = info_to_html (info);
    total_str = g_strdup_printf (data->template, info_str);

    gtk_html_write (data->html, data->stream, 
                    total_str, strlen (total_str));

    g_free (info_str);
    g_free (total_str);
}

static void
render_info (RCResolverContext *solution,
             RCPackage *package, int priority,
             GtkHTML *html, GtkHTMLStream *stream,
             const char *template)
{
    struct RenderInfoData data;

    data.html = html;
    data.stream = stream;
    data.template = template;

    rc_resolver_context_foreach_info (solution,
                                      package, priority,
                                      render_info_cb,
                                      &data);
}

/* ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** */

static void
render_packages(RCPackageSList *p, GtkHTML *html, 
                GtkHTMLStream *stream, GUIChannel *channel,
                gboolean removal, gboolean dependency_info,
                RCResolverContext *solution)
{
    RCPackageSList *packages;
    GSList *i;
    char *h;

    packages = g_slist_copy(p);
    if (gui_prefs_get_pretty_names())
        packages = rc_package_slist_sort_by_pretty_name(packages);
    else
        packages = rc_package_slist_sort_by_name(packages);

    i = packages;
    while (i) {
        RCPackage *pkg = i->data;
        char *old_version = NULL;
        char *new_version;
        char *size;

        if (!removal) {
            RCPackageUpdate *upd = pkg->history->data;
            RCPackage *old_pkg = rc_package_new();

            old_pkg->spec.name = g_strdup(pkg->spec.name);
            old_pkg = rc_packman_query(appinfo->pman, old_pkg);
        
            if (old_pkg->installed) {
                old_version = rc_package_spec_version_to_str(
                    RC_PACKAGE_SPEC(old_pkg));
            }
            else {
                old_version = g_strdup("--");
            }
            rc_package_free(old_pkg);

            size = rc_gui_format_size_with_nbsp(upd->package_size);
        }
        else
            size = rc_gui_format_size_with_nbsp(pkg->installed_size);

        new_version = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(pkg));
        
        h = g_strdup_printf(
            "<tr valign=top>"
            "<td width=0>&nbsp;&nbsp;</td>"
            "<td><b>%s</b></td>"
            "<td>%s</td>"
            "<td>%s</td>"
            "%s%s%s"
            "<td></td>"
            "</tr>",
            rc_gui_get_pretty_name(pkg->spec.name), new_version, size,
            removal ? "" : "<td>", removal ? "" : old_version,
            removal ? "" : "</td>");
        gtk_html_write(html, stream, h, strlen(h));
        g_free(old_version);
        g_free(new_version);
        g_free(size);
        g_free(h);

        if (dependency_info && solution) {

            render_info (solution,
                         pkg, RC_RESOLVER_INFO_PRIORITY_USER,
                         html, stream,
                         "<tr valign=\"top\">"
                         "<td width=\"0\">&nbsp;&nbsp;</td>"
                         "<td colspan=5>&nbsp;&nbsp;&nbsp;%s</td>"
                         "</tr>");
        }

        if (pkg->channel) {
            GSList *s;
            GUIChannel *chan = NULL;

            s = appinfo->channel_list;
            while (s) {
                GUIChannel *c = s->data;

                if (c->rcchannel.id == pkg->channel->id) {
                    chan = c;
                    break;
                }

                s = s->next;
            }

            if ((channel && chan != channel) || (chan && !chan->subscribed)) {
                h =
                    "<tr><td>&nbsp</td><td colspan=6>"
                    "<center><table width=95%><tr><td width=0%>"
                    "<img src=\"file://"SHAREDIR"/dep-warning.png\""
                    "     height=16>"
                    "</td><td width=100%><font size=-1>";
                gtk_html_write(html, stream, h, strlen(h));
                    
                if (chan->subscribed) {
                    h = g_strdup_printf
                        (U_("<!--html-->This package will be pulled in "
                            "from the %s channel."),
                         RCCHANNEL(chan)->name);
                } else {
                    h = g_strdup_printf
                        (U_("<!--html-->This package will be pulled in "
                            "from the <b>unsubscribed</b> %s channel."),
                         RCCHANNEL(chan)->name);
                }

                gtk_html_write(html, stream, h, strlen(h));
                g_free(h);
                    
                h = "</font></td></tr></table></center></td></tr>";
                gtk_html_write(html, stream, h, strlen(h));
            }
        }
        
        i = i->next;
    }

    g_slist_free(packages);

    h = "<tr><td colspan=7>&nbsp;</td></tr>";
    gtk_html_write(html, stream, h, strlen(h));
} /* render_packages */

gboolean
rc_gui_render_dependencies(RCPackageSList *requested,
                           RCPackageSList *removed,
                           RCPackageSList *added,
                           RCResolverContext *solution,
                           GUIChannel *channel,
                           gboolean removal)
{
    GtkHTML *html;
    GtkHTMLStream *stream;
    char *h;
    gchar *tmp;
    guint count;

    /* Do not allow select all/none or subscribe/unsubscribe here */
    rc_gui_sensitize_select_items(SELECT_NEITHER);
    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);

    html = GTK_HTML(
        gtk_object_get_user_data(GTK_OBJECT(GET_WIDGET("verify_head"))));
    stream = gtk_html_begin(html);
    h = g_strdup_printf (
        "<html><body bgcolor=\"#000000\" marginwidth=0 marginheight=1>"
        "<table cellspacing=0 cellpadding=0 width=\"100%%\" border=0>"
        "<tr>"
        "<td align=left><img src=\"file://"SHAREDIR"/download.png\">"
        "</td>"
        "<td align=left width=\"100%%\">"
        "<font size=\"+2\" color=\"#ffffff\">"
        "<b>%s</b></font>" /* "Package Dependencies" */
        "</td></tr></table>"
        "</body></html>",
        U_("<!--html-->Package Dependencies"));
    gtk_html_write(html, stream, h, strlen(h));
    g_free (h);
    gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

    html = GTK_HTML(
        gtk_object_get_user_data(GTK_OBJECT(GET_WIDGET("verify_html"))));
    stream = gtk_html_begin(html);

    if (!requested && !removed && !added) {
        h = g_strdup_printf (
            "<html><body bgcolor=#ffffff text=#000000 marginwidth=0"
            "            marginheight=0>"
            "<table width=100%% cellspacing=0 cellpadding=3 border=0>"
            "<tr>"
            "<td><img src=\"file://"SHAREDIR"/progress-verify.png\"></td>"
            "<td>"
            "<br>%s<br><br></td>"
            "</tr>"
            "</table></body></html>",
            U_("<!--html-->Your system's package database is valid!<br><br>"
               "Click the Previous button to go back."));
        gtk_html_write(html, stream, h, strlen(h));
        gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

        return TRUE;
    }

    h =
        "<html><body bgcolor=#ffffff text=#000000 marginwidth=0"
        "            marginheight=0>"
        "<table width=100% cellspacing=0 cellpadding=1 border=0>";
    gtk_html_write(html, stream, h, strlen(h));

    /* Make a little summary at the top, only if we're doing two or more things */
    count = 0;
    if (requested) count++;
    if (removed) count++;
    if (added) count++;

    if (count >= 2) {
        h =
            "<tr>"
            "<td width=0>&nbsp;&nbsp;"
            "</td><td width=\"100%\">";
        gtk_html_write(html, stream, h, strlen(h));

        if (requested) {
            int cnt = g_slist_length (requested);

            if (cnt == 1) {
                h = (char *) U_("<b>1</b> Requested Package");
                gtk_html_write (html, stream, h, strlen(h));
            } else {
                h = g_strdup_printf(U_("<b>%d</b> Requested Packages"), cnt);
                gtk_html_write (html, stream, h, strlen(h));
                g_free (h);
            }

            if (removed || added) {
                h = "&nbsp;|&nbsp;";
                gtk_html_write (html, stream, h, strlen (h));
            }
        }

        if (removed) {
            int cnt = g_slist_length (removed);

            if (cnt == 1) {
                h = (char *) U_("<b>1</b> Necessary Removal");
                gtk_html_write (html, stream, h, strlen(h));
            } else {
                h = g_strdup_printf (U_("<b>%d</b> Necessary Removals"), cnt);
                gtk_html_write (html, stream, h, strlen(h));
                g_free (h);
            }

            if (added) {
                h = "&nbsp;|&nbsp;";
                gtk_html_write (html, stream, h, strlen (h));
            }
        }

        if (added) {
            int cnt = g_slist_length (added);

            if (cnt == 1) {
                h = (char *) U_("<b>1</b> Necessary Installation");
                gtk_html_write (html, stream, h, strlen(h));
            } else {
                h = g_strdup_printf (U_("<b>%d</b> Necessary Installations"),
                                     cnt);
                gtk_html_write (html, stream, h, strlen(h));
                g_free (h);
            }

        }
        h = "</td></tr>";
        gtk_html_write (html, stream, h, strlen(h));
    }

    /* Now do a disk space summary */
    {
        gint download_bytes_needed = 0;
        gint installed_size_change = 0;
        gint working_size = 0;
        GSList *iter;

        if (!removal) {
            iter = requested;
            while (iter) {
                RCPackage *p = (RCPackage *) iter->data;
                RCPackageUpdate *up = (RCPackageUpdate *) p->history->data;

                download_bytes_needed += up->package_size;
                installed_size_change += up->installed_size;
                working_size += up->installed_size;

                p = rc_package_spec_slist_find_name (appinfo->system_packages,
                                                     p->spec.name);
                if (p) {
                    installed_size_change -= p->installed_size;
                }
                iter = iter->next;
            }
        } else {
            iter = requested;
            while (iter) {
                RCPackage *p = (RCPackage *) iter->data;
                installed_size_change -= p->installed_size;
                iter = iter->next;
            }
        }

        iter = added;
        while (iter) {
            RCPackage *p = (RCPackage *) iter->data;
            RCPackageUpdate *up = (RCPackageUpdate *) p->history->data;

            download_bytes_needed += up->package_size;
            installed_size_change += up->installed_size;
            working_size += up->installed_size;

            p = rc_package_spec_slist_find_name (appinfo->system_packages,
                                                 p->spec.name);
            if (p) {
                installed_size_change -= p->installed_size;
            }

            iter = iter->next;
        }

        iter = removed;
        while (iter) {
            RCPackage *p = (RCPackage *) iter->data;
            installed_size_change -= p->installed_size;

            iter = iter->next;
        }

        h = "<tr><td width=0>&nbsp;&nbsp;</td><td>";
        gtk_html_write (html, stream, h, strlen (h));

        if (download_bytes_needed) {
            gchar *tmps = rc_gui_format_size_with_nbsp(download_bytes_needed);
            tmp = g_strdup_printf(
                U_("<!--html--><b>%s</b> will need to be downloaded. "), tmps);
            g_free (tmps);
            gtk_html_write (html, stream, tmp, strlen (tmp));
            g_free (tmp);
        }

        if (installed_size_change > 0) {
            gchar *tmps = rc_gui_format_size_with_nbsp(installed_size_change);
            tmp = g_strdup_printf(
                U_("<!--html-->After operations, <b>%s</b> of disk space "
                   "will be used."), tmps);
        } else if (installed_size_change < 0) {
            gchar *tmps = rc_gui_format_size_with_nbsp(
                - installed_size_change);
            tmp = g_strdup_printf(
                U_("<!--html-->After operations, <b>%s</b> of disk space "
                   "will be freed."), tmps);
        } else {
            tmp = g_strdup(
                U_("<!--html-->After operations, no additional space "
                   "will be freed or used."));
        }

        gtk_html_write (html, stream, tmp, strlen (tmp));
        g_free (tmp);

        h = "</td></tr>";
        gtk_html_write (html, stream, h, strlen (h));

        {
#ifdef HAVE_STATVFS
            struct statvfs stat, tmpstat, optstat;
#else
            struct statfs stat, tmpstat, optstat;
#endif
            guint32 bs, tmpbs, optbs;
            char *cachedir, *tmpdir;
            gboolean opt;

            gui_prefs_get_cache_info(&cachedir, NULL, NULL, NULL);
            tmpdir = getenv ("TMPDIR");
            if (!tmpdir)
                tmpdir = "/tmp";
#ifdef HAVE_STATVFS
            statvfs(cachedir, &stat);
            bs = stat.f_frsize;
            statvfs(tmpdir, &tmpstat);
            tmpbs = tmpstat.f_frsize;
#else
            statfs(cachedir, &stat);
            bs = stat.f_bsize;
            statfs(tmpdir, &tmpstat);
            tmpbs = tmpstat.f_bsize;
#endif
            g_free(cachedir);
            if ((download_bytes_needed / bs) + 1 > stat.f_bavail ||
                (download_bytes_needed / tmpbs) + 1 > tmpstat.f_bavail) {
                char *m = g_strdup_printf(
                    "%s"
                    "<ul>"
                    "<li>%s"
                    "<li>%s"
                    "<li>%s"
                    "</ul>",
                    U_("<!--html-->You have insufficient disk space to "
                       "download the requested packages. You must free up "
                       "some disk space before you can continue. Some "
                       "options include:"),
                    U_("<!--html-->Removing some packages from your system"),
                    U_("Changing your <a href=\"updater:prefs_cache\">cache "
                       "settings</a>."),
                    U_("<a href=\"updater:cache_clear\">Clearing your "
                       "cache</a>."));
                gboolean retry;

                retry = rc_gui_error_box(m, RC_ERROR_RETRY);
                gtk_html_end(html, stream, GTK_HTML_STREAM_OK);
                if (retry) {
                    return rc_gui_render_dependencies (requested, removed, added, NULL, channel, removal);
                }
                else
                    return FALSE;
            }

            /* FIXME: This is totally not the right way to do this. */
#ifdef HAVE_STATVFS
            statvfs("/usr", &stat);
            bs = stat.f_frsize;
#else
            statfs("/usr", &stat);
            bs = stat.f_bsize;
#endif
            optbs = 0;
            opt = g_file_exists("/opt");
            if (opt) {
#ifdef HAVE_STATVFS
                statvfs("/opt", &optstat);
                optbs = stat.f_frsize;
#else
                statfs("/opt", &optstat);
                optbs = stat.f_bsize;
#endif
            }

            if (working_size / bs > stat.f_bavail ||
                (opt && working_size / optbs > stat.f_bavail)) {
                h = g_strdup_printf (
                    "<tr><td colspan=2><br></td></tr>"
                    "<tr><td width=0>"
                    "<img src=\"file://"SHAREDIR"/dep-warning.png\" height=16>"
                    "</td><td>"
                    "<b>%s</b>"
                    "</td></tr>",
                    U_("<font color=#ff0000>Warning!</font> There "
                       "may not be sufficient disk space to install this "
                       "package, and installation of this package may fail. "
                       "You should free up some disk space before "
                       "continuing."));
                gtk_html_write(html, stream, h, strlen(h));
                g_free (h);
            }
        }
    }

    h = "</table><br><br>";
    gtk_html_write (html, stream, h, strlen (h));

    count = g_slist_length (requested);

    if (count == 1) {
        tmp = g_strdup_printf (U_("1&nbsp;Package"));
    } else {
        tmp = g_strdup_printf (U_("%d&nbsp;Packages"), count);
    }

    if (requested) {
        h = "<a name=\"requested\">";
        gtk_html_write(html, stream, h, strlen(h));

        h = rc_gui_render_yellow_bar(
            "#fcdc7b", "#000000", LEFT_CORNER_IMG,
            removal ? "file://"SHAREDIR"/remove-button.png" :
            "file://"SHAREDIR"/install-button.png",
            removal ? U_("Requested Removals") : U_("Requested Installations"),
            tmp, RIGHT_CORNER_IMG);
        gtk_html_write (html, stream, h, strlen (h));
        g_free (h);

        g_free (tmp);

        h = g_strdup_printf(
            "<table width=100%% cellspacing=0 cellpadding=1>"
            "<tr bgcolor=#e3e5e5>"
#if 0
            "<td width=0></td>"
#else
            "<td width=0>&nbsp;&nbsp;</td>"
#endif
            "<td width=\"40%%\">%s</td>" /* "Package" */
            "<td width=\"25%%\">%s</td>" /* "Version" */
            "<td width=\"%d%%\" colspan=\"%d\">Size</td>"
            "%s"
            "</tr>",
            U_("Package"),
            U_("Version"),
            removal ? 35 : 10,
            removal ? 2 : 1,
            removal ? "" : U_("<td width=25%>Old Version</td>"));
        gtk_html_write (html, stream, h, strlen (h));
        g_free(h);
        
        /* note that we pass NULL for extra_infos here on purpose, to 
         * avoid getting info printed about the requested packages
         */
        render_packages(
            requested, html, stream, NULL, removal, FALSE, solution);

        h = "</table>";
        gtk_html_write (html, stream, h, strlen (h));
    }

    if (removed) {
        h = "<a name=\"removed\">";
        gtk_html_write(html, stream, h, strlen(h));

        count = g_slist_length (removed);

        if (count == 1) {
            tmp = g_strdup_printf (U_("1&nbsp;Package"));
        } else {
            tmp = g_strdup_printf (U_("%d&nbsp;Packages"), count);
        }

        h = rc_gui_render_yellow_bar (
            "#ff0000", "#ffffff", LEFT_CORNER_RED,
            "file://"SHAREDIR"/remove-button.png",
            U_("Necessary Removals"), tmp, RIGHT_CORNER_RED);
        gtk_html_write (html, stream, h, strlen (h));
        g_free (h);

        g_free (tmp);

        h = g_strdup_printf (
            "<table width=100%% cellspacing=0 cellpadding=1>"
            "<tr bgcolor=#e3e5e5>"
#if 0
            "<td width=0></td>"
#else
            "<td width=0>&nbsp;&nbsp;</td>"
#endif
            "<td width=40%%>%s</td>" /* "Package" */
            "<td width=25%%>%s</td>" /* "Version" */
            "<td width=10%%>%s</td>" /* "Size" */
            "<td width=25%%>&nbsp;</td>"
            "</tr>",
            U_("Package"),
            U_("Version"),
            U_("Size"));
        gtk_html_write (html, stream, h, strlen (h));

        render_packages (removed, html, stream, NULL, TRUE, TRUE, solution);

        h = "</table>";
        gtk_html_write (html, stream, h, strlen (h));
    }

    if (added) {
        h = "<a name=\"installs\">";
        gtk_html_write(html, stream, h, strlen(h));

        count = g_slist_length (added);

        if (count == 1) {
            tmp = g_strdup_printf (U_("1&nbsp;Package"));
        } else {
            tmp = g_strdup_printf (U_("%d&nbsp;Packages"), count);
        }

        h = rc_gui_render_yellow_bar (
            "#fcdc7b", "#000000", LEFT_CORNER_IMG,
            "file://"SHAREDIR"/install-button.png",
            U_("Necessary Installations"), tmp, RIGHT_CORNER_IMG);
        gtk_html_write (html, stream, h, strlen (h));
        g_free (h);

        g_free (tmp);

        h = g_strdup_printf (
            "<table width=100%% cellspacing=0 cellpadding=1>"
            "<tr bgcolor=#e3e5e5>"
#if 0
            "<td width=0></td>"
#else
            "<td width=0>&nbsp;&nbsp;</td>"
#endif
            "<td width=40%%>%s</td>" /* "Package" */
            "<td width=25%%>%s</td>" /* "Version" */
            "<td width=10%%>%s</td>" /* "Size" */
            "<td width=25%%>%s</td>" /* "Old Version" */
            "</tr>",
            U_("Package"),
            U_("Version"),
            U_("Size"),
            U_("Old Version"));
        gtk_html_write(html, stream, h, strlen(h));

        render_packages (added, html, stream, channel, FALSE, TRUE, solution);

#if 0
        if (extra_infos) {
            GSList *iter;
            iter = added;
            while (iter) {
                RCPackage *ap = (RCPackage *) iter->data;
                RCPackageSList *dep_on;
                dep_on = rc_dep_find_extra_info (extra_infos, ap->spec.name);
                if (dep_on) {
                    GSList *it;
                    fprintf (stderr, "Package %s is needed by ", ap->spec.name);
                    for (it = dep_on; it; it = it->next)
                        fprintf (stderr, "%s ", ((RCPackage *) it->data)->spec.name);
                    fprintf (stderr, "\n");
                } else {
                    fprintf (stderr, "Package %s must be upgraded.\n", ap->spec.name);
                }
                iter = iter->next;
            }
        }
#endif

        h = "</table>";
        gtk_html_write (html, stream, h, strlen (h));
    }        

    h = "</body></html>";
    gtk_html_write(html, stream, h, strlen(h));

    gtk_html_end(html, stream, GTK_HTML_STREAM_OK);

    return TRUE;
} /* rc_gui_render_dependencies */

/*
 * rc_gui_render_package_line
 *
 * Renders package with the op being the URI.
 * sz is the size that should be displayed (either update or size
 * from spec)
 *
 * endrow should be TRUE if the final </TR> is to be
 * included or not (so that stuff can be added after the row)
 */
static gchar *
rc_gui_render_package_line (int count, RCPackage *package,
                            guint32 sz, gchar *op, gint flags)
{
    gchar *html;
    gchar *size;
    char *version;
    gboolean older = flags & RENDER_PACKAGE_FLAGS_OLDER;
    gboolean discolor = flags & RENDER_PACKAGE_FLAGS_DISCOLOR;

    version = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(package));
    size = rc_gui_format_size_with_nbsp(sz);
    if (!appinfo->not_root) {
        html = g_strdup_printf(
            "<tr valign=top %s>"
            "<td nowrap>"
            "<a href=\"updater:%s?%d,%p\">"
            "<img src=\"updater:check_nothing\" width=16 height=16 "
            "     border=no id=\"ckimg%d\">"
            "</a>"
            "<a href=\"updater:info?%p\">"
            "<img src=\"file://"SHAREDIR"/info.png\" width=16"
            "     height=16 border=no>"
            "</a>"
            "</td>"
            "<td><a href=\"updater:%s?%d,%p\"><font %s>"
            "%s</font></a>"
            "</td>"
            "<td>%s</td>"
            "<td>%s</td>"
            "</tr>",
            (discolor && count % 2) ? "bgcolor=\"#f0f2f2\"" : "",
            op, count, package, count, package,
            op, count, package,
            older ? "color=\"#777777\"" : "",
            rc_gui_get_pretty_name(package->spec.name), 
            version,
            size);
    }
    else {
        html = g_strdup_printf(
            "<tr valign=top %s>"
            "<td nowrap>"
            "<a href=\"updater:info?%p\">"
            "<img src=\"file://"SHAREDIR"/info.png\" width=16"
            "     height=16 border=no>"
            "</a>"
            "</td>"
            "<td><a href=\"updater:%s?%d,%p\"><font %s>"
            "%s</font></a>"
            "</td>"
            "<td>%s</td>"
            "<td>%s</td>"
            "</tr>",
            (discolor && count % 2) ? "bgcolor=\"#f0f2f2\"" : "",
            package,
            op, count, package,
            older ? "color=\"#777777\"" : "",
            rc_gui_get_pretty_name(package->spec.name), 
            version,
            size);
    }
    g_free(version);
    g_free (size);
    return html;
} /* rc_gui_render_package_line */

/*
 * rc_gui_render_info_html
 *
 * Renders the channel info page
 */
void
rc_gui_render_info_html(GUIChannel *channel)
{
    char *url;
    GtkHTMLStream *stream;
    const char *html_data;

    /* Hide the package info HTML page, just because it seems like the
       appropriate thing to do when showing the info page */
    gtk_widget_hide(GET_WIDGET("package_vbox"));

    /* Do not allow selecting of all/none here */
    rc_gui_sensitize_select_items(SELECT_NEITHER);

    stream = gtk_html_begin (GTK_HTML (appinfo->info_html));
    html_data = 
        U_("<html><body><b>Please wait, loading page...</b></body></html>");
    gtk_html_write (GTK_HTML (appinfo->info_html), stream, html_data,
                    strlen (html_data));
    gtk_html_end (GTK_HTML (appinfo->info_html), stream, GTK_HTML_STREAM_OK);

    rc_gui_set_action_html(channel, ACTION_INFO, 0);

    url = rc_gui_get_mirrored_url(
        channel->subscribed ? RCCHANNEL(channel)->subs_file :
        RCCHANNEL(channel)->unsubs_file, RCCHANNEL(channel));

    /* Set this to be the base */
    g_free (gtk_object_get_data (GTK_OBJECT (appinfo->info_html), "base_url"));
    gtk_object_set_data (GTK_OBJECT (appinfo->info_html), "base_url", url);
    rc_gui_html_download (appinfo->info_html, NULL, url);
} /* rc_gui_render_info_html */


/*
 * rc_gui_render_packagesets_html
 *
 * Renders page of package sets available
 * (still needs work)
 */
void
rc_gui_render_packagesets_html(GUIChannel *channel)
{
    char *html;
    GtkHTMLStream *stream;
    GSList *iter;

    stream = gtk_html_begin (appinfo->update_html);

    if (channel->rcchannel.package_sets == NULL) {
        /* No packages have been installed from this channel */
        html = g_strdup_printf (
            "<html><body marginwidth=8 marginheight=4 bgcolor=#ffffff>"
            /* "You have no packages..." */
            "<p>%s</p>"
            /* "You can visit... install software." */
            "%s<br>"
            /* View the about page */
            "<ul><li>%s"
            /* View the install page */
            "<li>%s"
            /* go back to summary */
            "<li>%s"
            "</ul>"
            "</body></html>",
            U_("You have no packages from this channel currently installed "
               "on your system."),
            U_("You can visit the channel's "
               "<a href=\"updater:info_page\">about page</a> to get more "
               "information about what software is available, or you can go "
               "directly to the <a href=\"updater:available_page\">install "
               "page</a> to install software."),
            U_("View the <a href=\"updater:info_page\">about page</a>."),
            U_("View the <a href=\"updater:available_page\">install "
               "page</a>."),
            U_("Return to the <a href=\"updater:summary\">Summary</a>."));
        gtk_html_write (appinfo->installed_html, stream, html, strlen (html));
        g_free (html);
        gtk_html_end (appinfo->installed_html, stream, GTK_HTML_STREAM_OK);
        return;
    }

    html = g_strdup_printf (
        "<html><body marginwidth=8 marginheight=4>"
        "<p>%s</p></body></html>",
        U_("You don't have any packages installed from this channel. "
           "The following pre-defined sets of packages are available, or "
           "you may select individual packages from the "
           "<a href=updater:available_page>Install page</a>."));
    gtk_html_write (appinfo->update_html, stream, html, strlen(html));

    iter = channel->rcchannel.package_sets;
    while (iter) {
        RCPackageSet *ps = (RCPackageSet *) iter->data;

        html = g_strdup_printf (
            "<a href=\"updater:install_set?%s\"<b>%s</b></a><br>"
            "<i>Size: %s</i><br>"
            "<p>%s</p>",
            ps->name,
            ps->name,
            "Calculate Me!",
            ps->description);
        gtk_html_write (appinfo->update_html, stream, html, strlen(html));
        g_free (html);
        iter = iter->next;
    }
    html = "</body></html>";
    gtk_html_write (appinfo->update_html, stream, html, strlen(html));
    gtk_html_end(appinfo->update_html, stream, GTK_HTML_STREAM_OK);

} /* render_packagesets_html */

/*
 * rc_gui_render_updates_html
 *
 * Renders page of per-channel updates
 */
void
rc_gui_render_updates_html(GUIChannel *channel)
{
    char *html;
    RCPackageSList **imp;
    int count = 0;
    RCPackageSList **packages = g_new0(RCPackageSList *, 1);
    int total_size = 0;
    char *total_size_string;
    char *n;
    GtkHTMLStream *stream;
    GtkWidget *w;

    w = gtk_entry_new();
    gtk_object_set_data(GTK_OBJECT(appinfo->update_html), "searchentry", w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(search_box_activated), 
        NULL);
    gtk_widget_show(w);

    appinfo->current_page = PAGE_UPDATES;
    rc_gui_sensitize_select_items(SELECT_ITEM_ALL | SELECT_ITEM_NONE);
    
    count = rc_gui_split_by_package_importance(
        channel->available_updates, channel->installed_packages, &imp, 
        &total_size);

    stream = gtk_html_begin(appinfo->update_html);
    if (!count) {
        g_free(imp);

        rc_gui_sensitize_select_items(SELECT_NEITHER);

        html = "<html><body marginwidth=1 marginheight=1>";
        gtk_html_write(appinfo->update_html, stream, html, strlen(html));

        if (g_slist_length (channel->installed_packages) != 0) {
            html = g_strdup_printf (
                "<html><body bgcolor=\"#ffffff\" text=\"#000000\""
                "            marginwidth=8 marginheight=4>"
                /* "All of the software .. is up to date" */
                "<p>%s</p><br>"
                "<ul>"
                /* "To install new software..." */
                "<li>%s"
                /* "To remove already installed..." */
                "<li>%s"
                /* "Unsubscribe from this channel." */
                "<li>%s"
                /* "Return to the summary" */
                "<li>%s"
                "</ul>"
                "</html></body>",
                U_("All of the software from this channel is up-to-date."),
                U_("To install new software from this channel, visit the "
                   "<a href=\"updater:available_page\">install page</a>."),
                U_("To remove already installed software from this channel, "
                   "visit the <a href=\"updater:installed_page\">remove "
                   "page</a>."),
                U_("<a href=\"updater:unsubscribe\">Unsubscribe</a> from this "
                   "channel."),
                U_("Return to the <a href=\"updater:summary\">Summary</a>."));

            gtk_html_write(appinfo->update_html, stream, html, strlen(html));
            gtk_html_end(appinfo->update_html, stream, GTK_HTML_STREAM_OK);
        } else {
            gtk_html_end(appinfo->update_html, stream, GTK_HTML_STREAM_OK);
            rc_gui_render_packagesets_html(channel);
        }
        rc_gui_set_action_html(channel, ACTION_UPDATE, 1);
        return;
    }

    rc_gui_set_action_html(channel, ACTION_UPDATE, 0);

    total_size_string = rc_gui_format_size(total_size);
    n = rc_gui_number_to_word(count);

    html = "<html><body bgcolor=#ffffff text=#000000 "
        "            link=\"#000000\""
        "            marginwidth=0"
        "            marginheight=0>"
        "<table width=100%% cellspacing=0 cellpadding=1 border=0>"
        "<tr><td>";
    gtk_html_write (appinfo->update_html, stream, html, strlen (html));

    if (count == 1) {
        html = g_strdup_printf (
            U_("There is <b>1 update</b> available in this "
               "channel, totalling <b>%s</b> of data to be downloaded."),
            total_size_string);
    } else {
        html = g_strdup_printf (
            U_("There are <b>%s updates</b> available in this "
               "channel, totalling <b>%s</b> of data to be downloaded."),
            n,
            total_size_string);
    }
    gtk_html_write (appinfo->update_html, stream, html, strlen (html));
    g_free (html);

    if (appinfo->not_root) {
        html = (char *) 
            U_("  You will not be able to update your system unless you run "
               "Red Carpet as root.");
        gtk_html_write (appinfo->update_html, stream, html, strlen (html));
    }

    g_free(n);
    g_free(total_size_string);

    html = "<br><br></td></tr></table>";
    gtk_html_write(appinfo->update_html, stream, html, strlen(html));

    {
        GSList *iter;
        int cur_imp;
        int has_updates[RC_IMPORTANCE_LAST] = { 0 };

        for (cur_imp = 0; cur_imp < RC_IMPORTANCE_LAST; cur_imp++) {
            if (imp[cur_imp]) {
                has_updates[cur_imp] += g_slist_length (imp[cur_imp]);
            }
        }

        count = 0;

        for (cur_imp = 0; cur_imp < RC_IMPORTANCE_LAST; cur_imp++) {
            const char *importance_names[RC_IMPORTANCE_LAST] = {
                U_("Essential Updates"),
                U_("Urgent Updates"),
                U_("Suggested Updates"),
                U_("Feature Enhancements"),
                U_("Minor Fixes/Updates")
            };

            if (has_updates[cur_imp]) {
                GSList *pkgs;
                gchar *tmp;
                int update_desc_count;

                if (has_updates[cur_imp] == 1) {
                    tmp = g_strdup (U_("1 update"));
                } else {
                    tmp = g_strdup_printf(
                        U_("%d updates"), has_updates[cur_imp]);
                }
                html = rc_gui_render_yellow_bar (
                    "#fcdc7b", "#000000", LEFT_CORNER_IMG, NULL,
                    importance_names[cur_imp], tmp, RIGHT_CORNER_IMG);
                gtk_html_write(
                    appinfo->update_html, stream, html, strlen(html));
                g_free(html);
                g_free (tmp);

                html = g_strdup_printf (
                    "<table width=\"100%%\" cellspacing=0 cellpadding=1>"
                    "<tr bgcolor=\"#e3e5e5\">"
                    "<td width=0 nowrap>&nbsp;</td>"
                    "<td width=45%%>%s</td>" /*"Package"*/
                    "<td width=40%%>%s</td>" /*"Version"*/
                    "<td width=15%%>%s</td>" /*"Size"*/
                    "</tr>",
                    U_("Package"),
                    U_("Version"),
                    U_("Size"));
                gtk_html_write(
                    appinfo->update_html, stream, html, strlen(html));
                g_free (html);

                /* Now fill out the list of packages */
                pkgs = imp[cur_imp];
                if (gui_prefs_get_pretty_names())
                    pkgs = rc_package_slist_sort_by_pretty_name (pkgs);
                else
                    pkgs = rc_package_slist_sort_by_name(pkgs);

                while (pkgs) {
                    RCPackage *pkg = (RCPackage *) pkgs->data;
                    RCPackageUpdate *update = (RCPackageUpdate *) 
                        pkg->history->data;
                    RCPackage *sys_pkg = g_new0 (RCPackage, 1);
                    gchar *sz_str;

                    sz_str = rc_gui_format_size_with_nbsp(
                        update->package_size);

                    html = rc_gui_render_package_line(
                        count, pkg, update->package_size,
                        "install_toggle", RENDER_PACKAGE_FLAGS_DISCOLOR);

                    gtk_html_write(
                        appinfo->update_html, stream, html, strlen(html));
                    g_free (html);
                    g_free (sz_str);

                    html = g_strdup_printf(
                        "<tr %s><td>&nbsp;</td><td colspan=4>",
                        count % 2 ? "bgcolor=\"#f0f2f2\"" : "");
                    gtk_html_write(
                        appinfo->update_html, stream, html, strlen(html));
                    g_free (html);

                    sys_pkg->spec.name = g_strdup (pkg->spec.name);
                    sys_pkg = rc_packman_query (appinfo->pman, sys_pkg);

                    update_desc_count = 0;
                    iter = pkg->history;
                    while (iter) {
                        /* Iterate through the updates in a package in an
                           importance level.  Let's only count the five
                           most recent, not including urgent or necessary
                           updates. */
                        RCPackageUpdate *this_update = iter->data;

                        if (rc_packman_version_compare(
                                appinfo->pman,
                                RC_PACKAGE_SPEC(this_update), 
                                RC_PACKAGE_SPEC(sys_pkg)) > 0 &&
                            update_desc_count < 5) {

                            html = g_strdup_printf(
                                "<li>%s", this_update->description);
                            gtk_html_write(
                                appinfo->update_html, stream, html,
                                strlen(html));
                            g_free(html);
                            /* Lower importances are numerically greater.
                               This makes no sense. */
                            if (this_update->importance > RC_IMPORTANCE_URGENT)
                                update_desc_count++;
                        }
                        iter = iter->next;
                    }

                    if (update_desc_count == 5) {
                        html = "<li>(older updates not shown)";
                        gtk_html_write(
                            appinfo->update_html, stream, html, strlen(html));
                    }

                    rc_package_free (sys_pkg);

                    html = "</td></tr>";
                    gtk_html_write(
                        appinfo->update_html, stream, html, strlen(html));

                    count++;
                    pkgs = pkgs->next;
                }

                html = "<tr><td colspan=5>&nbsp;</td><tr></table>";
                gtk_html_write(
                    appinfo->update_html, stream, html, strlen(html));

                g_slist_free (imp[cur_imp]);
            } /* if has_updates */
        } /* importance # loop */

    }

    g_free(imp);

    html = "</body></html>";
    gtk_html_write(appinfo->update_html, stream, html, strlen(html));

    gtk_object_set_data(
        GTK_OBJECT(appinfo->update_html), "checkbox_count",
        GINT_TO_POINTER(count));

    if (gtk_object_get_data(
            GTK_OBJECT(appinfo->update_html), "packages_to_install")) {
        g_free(gtk_object_get_data(
                   GTK_OBJECT(appinfo->update_html), "packages_to_install"));
    }

    gtk_object_set_data(
        GTK_OBJECT(appinfo->update_html), "packages_to_install", packages);

    gtk_html_end(appinfo->update_html, stream, GTK_HTML_STREAM_OK);
} /* rc_gui_render_updates_html */

/*
 * rc_gui_render_installed_html
 *
 * Renders page of currently installed pkgs
 */
void
rc_gui_render_installed_html(GUIChannel *channel)
{
    char *html;
    int count = 0;
    RCPackageSList **remove = g_new0(RCPackageSList *, 1);
    RCPackageSList **cat;
    int j;
    GtkHTMLStream *stream;
    GtkWidget *w;

    w = gtk_entry_new();
    gtk_object_set_data(GTK_OBJECT(appinfo->installed_html), "searchentry", w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(search_box_activated),
        NULL);
    gtk_widget_show(w);

    appinfo->current_page = PAGE_INSTALLED;
    rc_gui_sensitize_select_items(SELECT_ITEM_NONE);

    stream = gtk_html_begin(appinfo->installed_html);

    if (g_slist_length (channel->installed_packages) == 0) {
        rc_gui_sensitize_select_items(SELECT_NEITHER);

        /* No packages have been installed from this channel */
        html = g_strdup_printf (
            "<html><body marginwidth=8 marginheight=4 bgcolor=#ffffff>"
            /* "You have no packages..." */
            "<p>%s</p>"
            /* "You can visit... install software." */
            "%s<br>"
            /* View the about page */
            "<ul><li>%s"
            /* View the install page */
            "<li>%s"
            /* "Unsubscribe from this channel" */
            "<li>%s"
            /* go back to summary */
            "<li>%s"
            "</ul>"
            "</body></html>",
            U_("You have no packages from this channel currently installed "
               "on your system."),
            U_("You can visit the channel's "
               "<a href=\"updater:info_page\">about page</a> to get more "
               "information about what software is available, or you can go "
               "directly to the "
               "<a href=\"updater:available_page\">install page</a> to "
               "install software."),
            U_("View the <a href=\"updater:info_page\">about page</a>."),
            U_("View the <a href=\"updater:available_page\">install "
               "page</a>."),
            U_("<a href=\"updater:unsubscribe\">Unsubscribe</a> from this "
               "channel."),
            U_("Return to the <a href=\"updater:summary\">Summary</a>."));
        gtk_html_write (appinfo->installed_html, stream, html, strlen (html));
        g_free (html);
        gtk_html_end (appinfo->installed_html, stream, GTK_HTML_STREAM_OK);
        return;
    }

    html =
        "<html><body bgcolor=#ffffff text=#000000 "
        "            link=\"#000000\""
        "            marginwidth=0"
        "            marginheight=0>"
        "<table cellspacing=0 cellpadding=1 width=100% border=0>"
        "<tr><td>";
    gtk_html_write (appinfo->installed_html, stream, html, strlen (html));
    html = (char *)
        U_("The following packages from this channel are currently "
           "installed on your system.");
    gtk_html_write (appinfo->installed_html, stream, html, strlen (html));

    if (appinfo->not_root) {
        html = (char *)
            U_("<!--html-->"
               "  To remove packages, you will have to run Red Carpet "
               "as root.");
        gtk_html_write (appinfo->installed_html, stream, html, strlen (html));
    }

    html =
        "<br><br></td></tr></table>";
    gtk_html_write (appinfo->installed_html, stream, html, strlen (html));

    rc_gui_split_by_package_category(channel->installed_packages, &cat);

    for (j = 0; j < RC_SECTION_LAST; j++) {
        /* Iterate through the categories */
        if (cat[j]) {
            GSList *i;

            html = rc_gui_render_yellow_bar (
                "#fcdc7b", "#000000", LEFT_CORNER_IMG,
                rc_gui_section_icon_name (j),
                rc_package_section_to_user_string (j), NULL, RIGHT_CORNER_IMG);
            gtk_html_write (
                appinfo->installed_html, stream, html, strlen (html));
            g_free (html);

            html = g_strdup_printf (
                "<table width=100%% cellspacing=0 cellpadding=1>"
                "<tr bgcolor=#e3e5e5>"
                "<td width=0 nowrap>&nbsp;</td>"
                "<td width=45%%>%s</td>" /*"Package"*/
                "<td width=40%%>%s</td>" /*"Version"*/
                "<td width=15%%>%s</td>" /*"Size"*/
                "</tr>",
                U_("Package"),
                U_("Version"),
                U_("Size"));
            gtk_html_write(
                appinfo->installed_html, stream, html, strlen(html));
            g_free (html);

            if (gui_prefs_get_pretty_names())
                cat[j] = rc_package_slist_sort_by_pretty_name (cat[j]);
            else
                cat[j] = rc_package_slist_sort_by_name(cat[j]);

            i = cat[j];
            while (i) {
                /* Iterate through the list of packages in this category */
                RCPackage *package = i->data;
                
                html = rc_gui_render_package_line(
                    count, package, package->installed_size,
                    "remove_toggle", RENDER_PACKAGE_FLAGS_NONE);

                gtk_html_write(
                    appinfo->installed_html, stream, html, strlen(html));
                g_free(html);
                
                count++;
                i = i->next;
            }
            html = "<tr><td colspan=4>&nbsp;</td></tr></table>";
            gtk_html_write(
                appinfo->installed_html, stream, html, strlen(html));
        }
    }
    gtk_object_set_data(
        GTK_OBJECT(appinfo->installed_html), "checkbox_count",
        GINT_TO_POINTER(count));
    if (gtk_object_get_data(
            GTK_OBJECT(appinfo->installed_html), "packages_to_remove")) {
        g_free(gtk_object_get_data(
            GTK_OBJECT(appinfo->installed_html), "packages_to_remove"));
    }
    gtk_object_set_data(
        GTK_OBJECT(appinfo->installed_html), "packages_to_remove", remove);

    
    html = "</body></html>";
    gtk_html_write(appinfo->installed_html, stream, html, strlen(html));
    gtk_html_end(appinfo->installed_html, stream, GTK_HTML_STREAM_OK);
} /* rc_gui_render_installed_html */

/*
 * rc_gui_render_available_html
 *
 * Renders page of currently uninstalled pkgs
 */
void
rc_gui_render_available_html(GUIChannel *channel)
{
    char *html;
    int count = 0;
    RCPackageSList **install = g_new0(RCPackageSList *, 1);
    RCPackageSList **cat;
    int j;
    GtkHTMLStream *stream;
    GtkWidget *w;

    w = gtk_entry_new();
    gtk_object_set_data(GTK_OBJECT(appinfo->available_html), "searchentry", w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(search_box_activated),
        NULL);
    gtk_widget_show(w);
   
    appinfo->current_page = PAGE_AVAILABLE;
    rc_gui_sensitize_select_items(SELECT_ITEM_NONE | SELECT_ITEM_ALL);

    stream = gtk_html_begin(appinfo->available_html);

    if (g_slist_length (channel->uninstalled_packages) == 0
#ifndef BROKEN_OLDER
        || g_slist_length (channel->uninstalled_packages) == g_slist_length (channel->older_packages)
#endif
        )
    {
        rc_gui_sensitize_select_items(SELECT_NEITHER);

        /* All packages have been installed from this channel */
        html = 
            "<html><body marginwidth=8 marginheight=4 bgcolor=#ffffff>";
        gtk_html_write(appinfo->available_html, stream, html, strlen(html));
        html = (char *)
            U_("<p>All packages available in this channel are already "
               "installed on your system.</p>");
        gtk_html_write(appinfo->available_html, stream, html, strlen(html));

        if (channel->available_updates) {
            html = (char *)
                U_("<p>You can go to the "
                   "<a href=\"updater:update_page\">Update page</a> to view "
                   "available updates for your software in this channel, or "
                   "return to the <a href=\"updater:summary\">Summary</a> to "
                   "view all available updates.</p>");
        } else {
            html = (char *)
                U_("<p>You can go to the "
                   "<a href=\"updater:summary\">Summary</a> to view all "
                   "available updates for your system.</p>");
        }
        gtk_html_write(appinfo->available_html, stream, html, strlen(html));

        html = "</body></html>";
        gtk_html_end(appinfo->available_html, stream, GTK_HTML_STREAM_OK);
        return;
    }

    html =
        "<html><body bgcolor=\"#ffffff\" text=\"#000000\" "
        "            link=\"#000000\""
        "            marginwidth=0"
        "            marginheight=0>"
        "<table cellspacing=0 cellpadding=1 width=\"100%\" border=0>"
        "<tr><td>";
    gtk_html_write (appinfo->available_html, stream, html, strlen (html));

    html = (char *)
        U_("<!--html-->"
           "The following packages from this channel are available "
           "for installation.");
    gtk_html_write (appinfo->available_html, stream, html, strlen (html));

#ifdef BROKEN_OLDER
    html =
        U_(" Package names that are in <font color=\"#777777\">gray</font> indicate that a newer version of this package is already installed.");
    gtk_html_write (appinfo->available_html, stream, html, strlen (html));
#endif

    if (appinfo->not_root) {
        html = (char *)
            U_("<!--html-->"
               "  To install packages, you will have to run Red Carpet "
               "as root.");
        gtk_html_write (appinfo->available_html, stream, html, strlen (html));
    }

    html =
        "<br><br></td></tr></table>";
    gtk_html_write (appinfo->available_html, stream, html, strlen (html));

    rc_gui_split_by_package_category(channel->uninstalled_packages, &cat);

    for (j = 0; j < RC_SECTION_LAST; j++) {
        /* Iterate through the categories */
        if (cat[j]) {
            GSList *i;

            html = rc_gui_render_yellow_bar (
                "#fcdc7b", "#000000",
                LEFT_CORNER_IMG, rc_gui_section_icon_name (j),
                rc_package_section_to_user_string (j), NULL, RIGHT_CORNER_IMG);
            gtk_html_write(
                appinfo->available_html, stream, html, strlen(html));
            g_free(html);

            html = g_strdup_printf (
                "<table width=100%% cellspacing=0 cellpadding=1>"
                "<tr bgcolor=#e3e5e5>"
                "<td width=0 nowrap>&nbsp;</td>"
                "<td width=45%%>%s</td>" /*"Package"*/
                "<td width=40%%>%s</td>" /*"Version"*/
                "<td width=15%%>%s</td>" /*"Size"*/
                "</tr>",
                U_("Package"),
                U_("Version"),
                U_("Size"));
            gtk_html_write(
                appinfo->available_html, stream, html, strlen (html));
            g_free (html);

            if (gui_prefs_get_pretty_names())
                cat[j] = rc_package_slist_sort_by_pretty_name (cat[j]);
            else
                cat[j] = rc_package_slist_sort_by_name(cat[j]);

            i = cat[j];
            while (i) {
                /* Iterate through the list of packages in this category */
                RCPackage *package = i->data;
                /* The update we care about is the first one in the list, which
                   is always the most recent */
                RCPackageUpdate *update = package->history->data;

                if (rc_package_spec_slist_find_name
                    (channel->older_packages, package->spec.name))
                {
#ifdef BROKEN_OLDER
                    /* For now, skip older packages, until we fully do downgrades */
                    html = rc_gui_render_package_line(
                        count, package, update->package_size,
                        "install_toggle", RENDER_PACKAGE_FLAGS_OLDER);
#else
                    i = i->next;
                    continue;
#endif
                } else {
                    html = rc_gui_render_package_line(
                        count, package, update->package_size,
                        "install_toggle", RENDER_PACKAGE_FLAGS_NONE);
                }

                gtk_html_write(
                    appinfo->available_html, stream, html, strlen(html));
                g_free(html);
                
                count++;
                i = i->next;
            }
            html = "<tr><td colspan=5>&nbsp;</td></tr></table>";
            gtk_html_write(
                appinfo->available_html, stream, html, strlen(html));
        }
    }
    gtk_object_set_data(
        GTK_OBJECT(appinfo->available_html), "checkbox_count",
        GINT_TO_POINTER(count));
    if (gtk_object_get_data(
            GTK_OBJECT(appinfo->available_html), "packages_to_install")) {
        g_free(gtk_object_get_data(
            GTK_OBJECT(appinfo->available_html), "packages_to_install"));
    }
    gtk_object_set_data(
        GTK_OBJECT(appinfo->available_html), "packages_to_install", install);
    
    html = "</body></html>";
    gtk_html_write(appinfo->available_html, stream, html, strlen(html));
    gtk_html_end(appinfo->available_html, stream, GTK_HTML_STREAM_OK);
} /* rc_gui_render_available_html */

void
rc_gui_render_package_info_html(RCPackage *package)
{
    RCPackageUpdate *update = package->history ? 
        package->history->data : NULL;
    RCPackage *cpackage;
    GtkHTMLStream *stream;
    char *html;
    static RCPackage *currently_shown_package = NULL;
    GtkHTML *close;
    gboolean installed;
    char *version, *cversion;

    if (currently_shown_package && currently_shown_package == package &&
        GTK_WIDGET_VISIBLE(GET_WIDGET("package_vbox"))) {
        currently_shown_package = NULL;
        gtk_widget_hide (GET_WIDGET("package_vbox"));
        return;
    }

    currently_shown_package = package;

    cpackage = rc_package_new();
    cpackage->spec.name = g_strdup(package->spec.name);
    cpackage = rc_packman_query(appinfo->pman, cpackage);

    installed = cpackage->installed;
    if (!installed) {
        rc_package_free(cpackage);
        cpackage = package;
    }
    gtk_widget_show_all(GET_WIDGET("package_vbox"));

    close = GTK_HTML(gtk_object_get_user_data(
        GTK_OBJECT(GET_WIDGET("close_html"))));
    stream = gtk_html_begin(close);
    html = 
        "<html><body bgcolor=#000000 marginwidth=0 marginheight=0>"
        "<p align=left><a href=\"updater:close\">"
        "<img src=\"file://"SHAREDIR"/close.png\" width=15 height=15"
        "     border=no></a></p></body></html>";
    gtk_html_write(close, stream, html, strlen(html));
    gtk_html_end(close, stream, GTK_HTML_STREAM_OK);

    version = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(package));
    cversion = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(cpackage));

    stream = gtk_html_begin(appinfo->package_html);
    html = g_strdup_printf(
        "<table cellspacing=0 cellpadding=3 width=\"100%%\" border=0>"
        "<tr><td>"
        "<b>%s" /*"Name"*/ "</b> %s</td>"
        "<td align=right><b>%s" /*"Installed version:"*/ "</b></td>"
        "<td align=left>%s</td></tr>"
        "<tr><td><b>%s" /*"Size:"*/ "</b> %d %s" /*"bytes"*/ "</td>"
        "<td align=right><b>%s" /*"Requested Version:"*/ "</b></td>"
        "<td align=left>%s</td></tr>"
        "<tr><td colspan=3><b>%s" /*"Summary:"*/ "</b> %s</td></tr>"
        "<tr><td colspan=3><b>%s" /*"Description:"*/ "</b></td></tr>"
        "<tr><td colspan=3><pre>%s</pre></td></tr></table>",
        U_("Name:"),
        package->spec.name, 
        U_("Installed Version:"),
        cpackage->installed ? cversion : "(none)",
        U_("Size:"),
        package->installed ? package->installed_size : update->package_size,
        U_("bytes"),
        U_("Requested Version:"),
        version,
        U_("Summary:"),
        package->summary,
        U_("Description:"),
        package->description);

    gtk_html_write(appinfo->package_html, stream, html, strlen(html));
    g_free(html);

    if (package->requires || package->provides || package->conflicts) {
        html = "<table cellspacing=0 cellpadding=3 width=\"100%%\" border=0>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
    }

    if (package->requires) {
        GSList *iter = package->requires;

        html = "<tr><td valign=top align=right width=\"1%%\"><b>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
        html = (char *) U_("Requires:");
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
        html = "</b></td><td>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));

        while (iter) {
            RCPackageDep *di = (RCPackageDep *) iter->data;
            const gchar *rel = NULL;

            if (di->relation != RC_RELATION_ANY)
                rel = rc_package_relation_to_string (di->relation, 2);

            html = di->spec.name;
            gtk_html_write(appinfo->package_html, stream, html, strlen(html));

            if (rel) {
                char *v = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(di));
                html = g_strdup_printf(" %s %s", rel, v);
                g_free(v);
                gtk_html_write(appinfo->package_html, stream, html, strlen(html));
                g_free (html);
            }
#if 0
            if (dep->next) {
                html = g_strdup_printf(" <i>%s</i> ", U_("or")); /* package OR package OR package .. */
                gtk_html_write(appinfo->package_html, stream, html, strlen(html));
                g_free (html);
            }
            dep = dep->next;
#endif

            html = "<br>";
            gtk_html_write(appinfo->package_html, stream, html, strlen(html));
            iter = iter->next;
        }

        html = "</td></tr>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
    }

    if (package->provides) {
        GSList *iter = package->provides;

        html = "<tr><td valign=top align=right width=\"1%%\"><b>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
        html = (char *) U_("Provides:");
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
        html = "</b></td><td>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));

        while (iter) {
            RCPackageDep *di = (RCPackageDep *) iter->data;
            const gchar *rel = NULL;

            if (di->relation != RC_RELATION_ANY)
                rel = rc_package_relation_to_string (di->relation, 2);

            html = di->spec.name;
            gtk_html_write(appinfo->package_html, stream, html, strlen(html));

            if (rel) {
                char *v = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(di));
                html = g_strdup_printf(" %s %s", rel, v);
                g_free(v);
                gtk_html_write(appinfo->package_html, stream, html, strlen(html));
                g_free (html);
            }
#if 0
                if (dep->next) {
                    html = g_strdup_printf (" <i>%s</i> ", U_("or")); /* package OR package OR package .. */
                    gtk_html_write(appinfo->package_html, stream, html, strlen(html));
                    g_free (html);
                }
                dep = dep->next;
#endif

            html = "<br>";
            gtk_html_write(appinfo->package_html, stream, html, strlen(html));
            iter = iter->next;
        }

        html = "</td></tr>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
    }

    if (package->conflicts) {
        GSList *iter = package->conflicts;

        html = "<tr><td valign=top align=right width=\"1%%\"><b>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
        html = (char *) U_("Conflicts:");
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
        html = "</b></td><td>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));

        while (iter) {
            RCPackageDep *di = (RCPackageDep *) iter->data;
            const gchar *rel = NULL;

            if (di->relation != RC_RELATION_ANY)
                rel = rc_package_relation_to_string (di->relation, 2);

            html = di->spec.name;
            gtk_html_write(appinfo->package_html, stream, html, strlen(html));

            if (rel) {
                char *v = rc_package_spec_version_to_str(RC_PACKAGE_SPEC(di));
                html = g_strdup_printf(" %s %s", rel, v);
                g_free(v);
                gtk_html_write(appinfo->package_html, stream, html, strlen(html));
                g_free (html);
            }
#if 0
            if (dep->next) {
                html = g_strdup_printf (" <i>%s</i> ", U_("or")); /* package OR package OR package .. */
                gtk_html_write(appinfo->package_html, stream, html, strlen(html));
                g_free (html);
            }
            dep = dep->next;
#endif

            html = "<br>";
            gtk_html_write(appinfo->package_html, stream, html, strlen(html));
            iter = iter->next;
        }

        html = "</td></tr>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
    }

    if (package->requires || package->provides || package->conflicts) {
        html = "</table>";
        gtk_html_write(appinfo->package_html, stream, html, strlen(html));
    }

    html = "</body></html>";
    gtk_html_write(appinfo->package_html, stream, html, strlen(html));

    if (installed) {
        rc_package_free(cpackage);
    }

    gtk_html_end(appinfo->package_html, stream, GTK_HTML_STREAM_OK);
} /* render_package_html */

void
rc_gui_set_action_html(GUIChannel *channel, ActionType type, int additional)
{
    GtkHTMLStream *stream;
    char *h, *h1, *h2;
    char *object_classid = NULL;
    gboolean nolink = FALSE;
    GtkWidget *w;
    char *icon_url;

    w = gtk_entry_new();
    gtk_object_set_data(GTK_OBJECT(appinfo->action_html), "searchentry", w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(search_box_activated),
        NULL);
    gtk_widget_show(w);

    gtk_widget_show(GET_WIDGET("exec_head_window"));
    gtk_widget_show(GET_WIDGET("action_html_window"));

    switch (type) {
    case ACTION_SUMMARY:
        if (!additional) {
            object_classid = g_strdup_printf(
                "rcbutton:%p",
                rc_gui_make_button_callback_struct(
                    NULL,
                    _(" Update Now! "),
                    "updater:install_all",
                    channel, TRUE));
        }
        else {
            nolink = TRUE;
        }
        break;
    case ACTION_NEWS:
    case ACTION_FEATURED:
    case ACTION_SUBSCRIBED:
    case ACTION_UNSUBSCRIBED:
    case ACTION_ABOUT:
        nolink = TRUE;
        break;
    case ACTION_UPDATE:
        if (!additional) {
            object_classid = g_strdup_printf(
                "rcbutton:%p",
                rc_gui_make_button_callback_struct(
                    NULL,
                    _(" Update Packages "),
                    "updater:update",
                    channel, FALSE));
        }
        else {
            nolink = TRUE;
        }
        break;
    case ACTION_INSTALL:
        object_classid = g_strdup_printf(
            "rcbutton:%p",
            rc_gui_make_button_callback_struct(
                NULL,
                _(" Install Packages "),
                "updater:install",
                channel, FALSE));
        break;
    case ACTION_REMOVE:
        object_classid = g_strdup_printf(
            "rcbutton:%p",
            rc_gui_make_button_callback_struct(
                NULL,
                _(" Remove Packages "),
                "updater:remove",
                channel, FALSE));
        break;
    case ACTION_PREFERENCES:
        nolink = TRUE;
        break;
    case ACTION_INFO:
        g_return_if_fail(channel);

        if (channel->subscribed) {
            object_classid = g_strdup_printf(
                "rcbutton:%p",
                rc_gui_make_button_callback_struct(
                    appinfo->info_html, 
                    _(" Unsubscribe "),
                    "updater:unsubscribe",
                    channel, TRUE));
        }
        else {
            object_classid = g_strdup_printf(
                "rcbutton:%p",
                rc_gui_make_button_callback_struct(
                    appinfo->info_html,
                    _(" Subscribe "),
                    "updater:maybe_subscribe",
                    channel, TRUE));
        }

        break;
    case ACTION_SEARCH:
        nolink = TRUE;
        break;
    default:
        g_assert_not_reached();
    }

    h1 = g_strdup_printf (
        "<table cellspacing=0 cellpadding=7 width=100%% border=0>"
        "<tr><td align=left>"
        "<font color=#ffffff><b>%s" /*"Keyword search"*/ "</b></font></td>"
        "<td align=left width=100%%>"
        "<object classid=\"searchentry\"></object>"
        "</td></tr></table>",
        U_("Keyword&nbsp;Search:"));

    if (nolink || (appinfo->not_root && type != ACTION_INFO)) {
        h2 = g_strdup("&nbsp;");
    }
    else {
        h2 = g_strdup_printf(
            "<object classid=\"%s\"></object>", object_classid);
    }

    h = g_strdup_printf(
        "<html><body marginwidth=0 marginheight=0 bgcolor=\"#000000\">"
        "<table width=\"100%%\" height=\"100%%\" cellspacing=0 cellpadding=0>"
        "<tr valign=center>"
        "<td align=left width=\"50%%\">%s</td>"
        "<td nowrap align=right width=\"100%%\">%s&nbsp;&nbsp;&nbsp;</td>"
        "</tr></table></body></html>", h1, h2);

    g_free(h1);
    g_free(h2);

    stream = gtk_html_begin(appinfo->action_html);
    gtk_html_write(appinfo->action_html, stream, h, strlen(h));
    g_free (h);
    gtk_html_end(appinfo->action_html, stream, GTK_HTML_STREAM_OK);

    /* Let's do the header! */

    stream = gtk_html_begin(appinfo->exec_head);
    h = 
        "<html><body bgcolor=\"#000000\" marginwidth=0 marginheight=0>"
        "<table cellspacing=0 cellpadding=1 width=\"100%\" border=0"
        "       bgcolor=\"#000000\">";

    gtk_html_write(appinfo->exec_head, stream, h, strlen(h));
    
    switch (type) {
    case ACTION_SUMMARY:
        h = g_strdup_printf(
            "<tr><td align=left>"
            "<img src=\"file://"SHAREDIR"/summary.png\">"
            "</td><td align=left width=\"100%%\">"
            "<font size=\"+2\" color=\"#ffffff\">"
            "<b>%s</b></font></td></tr></table></body></html>",
            U_("Summary"));
        rc_gui_sensitize_select_items(SELECT_NEITHER);
        break;
    case ACTION_NEWS:
        h = g_strdup_printf(
            "<tr><td align=left>"
            "<img src=\"file://"SHAREDIR"/news.png\">"
            "</td><td align=left width=\"100%%\">"
            "<font size=\"+2\" color=\"#ffffff\">"
            "<b>%s</b></font></td></tr></table></body></html>",
            U_("Red Carpet News"));
        rc_gui_sensitize_select_items(SELECT_NEITHER);
        break;
    case ACTION_FEATURED:
        h = g_strdup_printf(
            "<tr><td align=left>"
            "<img src=\"file://"SHAREDIR"/featured.png\">"
            "</td><td align=left width=\"100%%\">"
            "<font size=\"+2\" color=\"#ffffff\">"
            "<b>%s</b></font></td></tr></table></body></html>",
            U_("Featured Channels"));
        rc_gui_sensitize_select_items(SELECT_NEITHER);
        break;
    case ACTION_SUBSCRIBED:
        h = g_strdup_printf(
            "<tr><td align=left>"
            "<img src=\"file://"SHAREDIR"/subscribed.png\">"
            "</td><td align=left width=\"100%%\">"
            "<font size=\"+2\" color=\"#ffffff\">"
            "<b>%s</b></font></td></tr></table>"
            "</body></html>",
            U_("Subscribed Channels"));
        rc_gui_sensitize_select_items(SELECT_NEITHER);
        break;
    case ACTION_UNSUBSCRIBED:
        h = g_strdup_printf(
            "<tr><td align=left>"
            "<img src=\"file://"SHAREDIR"/unsubscribed.png\">"
            "</td><td align=left width=\"100%%\">"
            "<font size=\"+2\" color=\"#ffffff\">"
            "<b>%s</b></font></td></tr></table>"
            "</body></html>",
            U_("Unsubscribed Channels"));
        rc_gui_sensitize_select_items(SELECT_NEITHER);
        break;
    case ACTION_ABOUT:
        h = g_strdup_printf(
            "<td><td align=left>"
            "<img src=\"file://"SHAREDIR"/download.png\">"
            "</td><td align=left width=\"100%%\">"
            "<font size=\"+2\" color=\"#ffffff\">"
            "<b>%s</b></font></td></tr></table></body></html>",
            U_("About"));
        rc_gui_sensitize_select_items(SELECT_NEITHER);
        break;
    default:
        /* We're in a channel */
        icon_url = rc_gui_get_mirrored_url(
            RCCHANNEL(channel)->icon_file, RCCHANNEL(channel));
        if (channel->subscribed) {
            h = g_strdup_printf(
                "<tr>"
                "<td align=left rowspan=2 width=48>"
                "<img height=48 width=48 src=\"%s\"></td>"
                "<td align=left rowspan=2 width=80%%>"
                "<font size=\"+2\" color=\"#ffffff\"><b>%s</b></font></td>"
                "<td width=9%%>%s</td>"
                "<td width=9%%>%s</td>"
                "<td width=9%%>%s</td>"
                "<td width=9%%>%s</td>"
                "<td width=5%%>&nbsp;</td>"
                "</tr>"
                "<tr>"
                "<td width=9%%><font size=-1 color=\"#ffffff\">"
                "%s%s%s</font></td>" /* Update */
                "<td width=9%%><font size=-1 color=\"#ffffff\">"
                "%s%s%s</font></td>" /* Install */
                "<td width=9%%><font size=-1 color=\"#ffffff\">"
                "%s%s%s</font></td>" /* Remove */
                "<td width=9%%><font size=-1 color=\"#ffffff\">"
                "%s%s%s</font></td>" /* About */
                "</tr>"
                "</table>", icon_url,
                RCCHANNEL(channel)->name,
                
                "<a href=\"updater:update_page\">"
                "<img width=32 height=32 src=\"file://" SHAREDIR 
                "/updates-button.png\" border=no></a>",
                
                "<a href=\"updater:available_page\">"
                "<img width=32 height=32 src=\"file://" SHAREDIR 
                "/install-button.png\" border=no></a>",
                
                "<a href=\"updater:installed_page\">"
                "<img width=32 height=32 src=\"file://" SHAREDIR 
                "/remove-button.png\" border=no></a>",
                
                "<a href=\"updater:info_page\">"
                "<img width=32 height=32 src=\"file://" SHAREDIR 
                "/info-button.png\" border=no></a>",
                
                type == ACTION_UPDATE ? "<b>" : "",
                U_("Update"),
                type == ACTION_UPDATE ? "</b>" : "",

                type == ACTION_INSTALL ? "<b>" : "",
                U_("Install"),
                type == ACTION_INSTALL ? "</b>" : "",

                type == ACTION_REMOVE ? "<b>" : "",
                U_("Remove"),
                type == ACTION_REMOVE ? "</b>" : "",
                
                type == ACTION_INFO ? "<b>" : "",
                U_("About"),
                type == ACTION_INFO ? "</b>" : "");
        }
        else {
            h = g_strdup_printf(
                "<tr>"
                "<td align=left rowspan=2 width=48><img src=\"%s\"></td>"
                "<td align=left rowspan=2 width=100%%>"
                "<font size=\"+2\" color=\"#ffffff\"><b>%s</b></font></td>"
                "<td width=9%%>%s</td><td width=5%%>&nbsp;</td>"
                "</tr>"
                "<tr><td width=9%%><font size=-1 color=#ffffff>"
                "<b>%s</b></font></td></tr>" /* About */
                "</table>", icon_url, 
                RCCHANNEL(channel)->name, 
                "<img src=\"file://" SHAREDIR "/info-button.png\">",
                U_("About"));
        }
        g_free (icon_url);
        break;
    }
    gtk_html_write(appinfo->exec_head, stream, h, strlen(h));
    g_free(h);
    gtk_html_end(appinfo->exec_head, stream, GTK_HTML_STREAM_OK);
} /* rc_gui_set_action_html */

static GHashTable *rand_hash = NULL;

static gboolean
rand_helper (gpointer key, gpointer data_p, gpointer udata_p)
{
    char *data = (char *) data_p;
    char **udata = (char **) udata_p;
    udata[g_hash_table_size (rand_hash) - 1] = data;
    return TRUE;
} /* rand_helper */

static char **
randomize_authors (char **dest, char **src)
{   
    int i;
    g_return_val_if_fail (rand_hash == NULL, NULL);

    rand_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
    
    for (i=0; src[i]; i++)
        g_hash_table_insert (rand_hash, GINT_TO_POINTER (rand()), src[i]);

    g_hash_table_foreach_remove (rand_hash, rand_helper, dest);
    g_hash_table_destroy (rand_hash);
    rand_hash = NULL;

    return dest;
} /* randomize_authors */

void
rc_gui_show_about(void)
{
    GtkHTMLStream *stream;
    char *h;
    int i;
    char *authors[] = {
        "Frank Belew",
        "Jacob Berkman",
        "Peter Bowen",
        "Javier D&iacute;az",
        "Anna Dirks",
        "Tuomas Kuosmanen",
        "Jose Mercado",
        "Carlos Montero-Luque",
        "Ian Peters",
        "Joe Shaw",
        "Jakub Steiner",
        "Peter Teichman",
        "Jon Trowbridge",
        "Alberto Tubilla",
        "Vladimir Vukicevic",
        NULL
    };
    char *randauth[16] = { NULL };

    rc_gui_set_action_html(NULL, ACTION_ABOUT, 0);
    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);
    rc_gui_set_refresh_item(REFRESH_SUMMARY);
    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), FALSE);

    stream = gtk_html_begin(appinfo->exec_body);
    h =
        "<html><body bgcolor=#e3e5e5 text=#000000>"
        "<table cellspacing=0 cellpadding=4 width=100%>"
        "<tr><td valign=top>"
        "<img src=\"file://"SHAREDIR"/about-monkey.png\" width=250"
        "     height=250>"
        "<td align=right width=100%>"
        "<font size=+1><b>Ximian Red Carpet "VERSION"</b></font><br>"
        "<b><font color=#dba70d>Copyright 2000-2002 Ximian, Inc."
        "</font></b><br><br><b>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    h = (char *) U_("Credits");
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    h = "</b><br>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    randomize_authors((char **) &randauth, (char **) &authors);
    for (i = 0; randauth[i]; i++) {
        h = g_strdup_printf("%s<br>", randauth[i]);
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        g_free(h);
    }

    h = "</td></tr><tr><td align=justify colspan=2><p>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    h = (char *)
        U_("<p>Red Carpet is the Ximian "
           "package management system and software portal, allowing "
           "you to install, upgrade, and remove packages on your system. For "
           "more information, please visit the "
           "<a href=\"http://www.ximian.com/redcarpet/\">Red Carpet web "
           "site</a>.</p>"
           "<p>We welcome your feedback and bug reports, either through "
           "the <a href=\"http://bugzilla.ximian.com/\">Ximian "
           "Bugzilla</a>, or the Red Carpet mailing list at "
           "<a href=\"http://lists.ximian.com/mailman/listinfo/red-carpet\">"
           "red-carpet@ximian.com</a>.  Thank you!");
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    h = "</p></td></tr>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

#ifdef BUILD
    if (BUILD && *BUILD) {
        h = "<br> Build: " BUILD;
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    }
#endif

    h = "</font></td></tr></table></body></html>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    gtk_html_end(appinfo->exec_body, stream, GTK_HTML_STREAM_OK);

    rc_gui_set_page("minor_notebook", PAGE_SUMMARY);
} /* rc_gui_show_about */

void
rc_gui_show_news_page(void)
{
    GtkHTMLStream *stream;
    char *h;
    GSList *i;

    g_return_if_fail(appinfo->news);

    rc_gui_set_action_html(NULL, ACTION_NEWS, 0);
    rc_gui_set_refresh_item(REFRESH_NEWS);
    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);

    stream = gtk_html_begin(appinfo->exec_body);

    h = "<html><body bgcolor=\"#ffffff\" marginwidth=8 marginheight=4><p>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    h = (char *) U_("All links will open in an external browser window.");
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    h =
        "</p>"
        "<table width=\"100%\" columns=2>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    i = appinfo->news->items;
    while (i) {
        RDFItem *item = i->data;
        gchar *image_url = NULL;

        if (!image_url && item->icon) {
            image_url = item->icon;
        }

        if (!image_url && item->channel) {
            RCChannel *rcc;
            rcc = rc_channel_get_by_name(
                (RCChannelSList *) appinfo->channel_list, item->channel);
            if (rcc)
                image_url = rcc->icon_file;
        }

        if (!image_url && appinfo->news->icon) {
            image_url = appinfo->news->icon;
        }

        if (!image_url) {
            image_url = SHAREDIR"/red-carpet.png";
        }

        h = g_strdup_printf (
            "<tr><td valign=top width=18><img src=\"%s\"></td>"
            "<td width=\"100%%\">", image_url);
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        g_free(h);
            
        h = g_strdup_printf (
            "<b><a href=\"%s\">%s</a></b><br>",
            item->location, item->title);
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        g_free(h);

        if (item->datestr) {
            h = g_strdup_printf (
                "<i>%s</i><br>",
                item->datestr);
            gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
            g_free(h);
        }

        if (item->summary) {
            h = g_strdup_printf (
                "%s<br>",
                item->summary);
            gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
            g_free(h);
        }

        h = "<br></td></tr>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        i = i->next;
    }

    h = "</table></body></html>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    gtk_html_end(appinfo->exec_body, stream, GTK_HTML_STREAM_OK);

    rc_gui_set_page("minor_notebook", PAGE_SUMMARY);
} /* rc_gui_show_news_page */

void
rc_gui_show_featured_page(void)
{
    GtkHTMLStream *stream;
    char *h;
    GSList *i;
    gboolean first_col = TRUE;

    rc_gui_set_action_html(NULL, ACTION_FEATURED, 0);
    rc_gui_set_refresh_item(REFRESH_SUMMARY);
    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);

    stream = gtk_html_begin(appinfo->exec_body);

    h =
        "<html><body text=#000000 marginwidth=0 marginheight=0>"
        "<table width=100% cellpadding=3 cellspacing=0 border=0>"
        "<tr bgcolor=#e3e5e5><td colspan=4>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    h = (char *) U_("The following channels are our featured channels!");
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    h = "<br><br></td></tr>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    for (i = appinfo->channel_list; i; i = i->next) {
        GUIChannel *channel = i->data;
        char *icon_url;

        if (!channel->rcchannel.featured)
            continue;

        if (first_col) {
            h = "<tr>";
            gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        }

        icon_url = rc_gui_get_mirrored_url(
            RCCHANNEL(channel)->icon_file, RCCHANNEL(channel));

        h = g_strdup_printf(
            "<td width=38 valign=top><a href=\"updater:channel?%d\">"
            "<img height=36 width=36 src=\"%s\" border=no></a></td>"
            "<td align=left width=45%%>"
            "<b><a href=\"updater:channel?%d\">%s</a></b></td>",
            channel->rcchannel.id, icon_url,
            channel->rcchannel.id, channel->rcchannel.name);
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        g_free(h);

        if (!first_col) {
            h = "</tr>";
            gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        }

        first_col = !first_col;
    }

    if (!first_col) {
        h = "<td width=45% align=left colspan=2>&nbsp;</td></tr>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    }

    h = "</table></body></html>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    gtk_html_end(appinfo->exec_body, stream, GTK_HTML_STREAM_OK);

    rc_gui_set_page("minor_notebook", PAGE_SUMMARY);
} /* rc_gui_show_featured_page */

void
rc_gui_show_channel_page (gboolean subscribed)
{
    GtkHTMLStream *stream;
    gchar *html;
    GSList *iter;
    gboolean first_column = TRUE;
    gint num_displayed = 0;

    rc_gui_set_action_html (NULL, subscribed ? ACTION_SUBSCRIBED :
                            ACTION_UNSUBSCRIBED, 0);
    rc_gui_set_refresh_item (REFRESH_SUMMARY);
    rc_gui_html_sensitize_subscribe (FALSE);
    rc_gui_html_sensitize_unsubscribe (FALSE);

    stream = gtk_html_begin (appinfo->exec_body);

    html =
        "<html><body bgcolor=#ffffff text=#000000 marginwidth=0 "
        "            marginheight=0>"
        "<table width=100%% cellpadding=3 cellspacing=0 border=0>";
    gtk_html_write (appinfo->exec_body, stream, html, strlen (html));

    for (iter = appinfo->channel_list; iter; iter = iter->next) {
        GUIChannel *channel = (GUIChannel *)(iter->data);
        gchar *icon_url;

        if ((channel->subscribed && !subscribed) ||
            (!channel->subscribed && subscribed))
        {
            continue;
        }

        num_displayed++;

        if (first_column) {
            html = "<tr><td height=10 colspan=5></td></tr><tr>";
            gtk_html_write (appinfo->exec_body, stream, html, strlen (html));
        }

        icon_url = rc_gui_get_mirrored_url(
            RCCHANNEL (channel)->icon_file, RCCHANNEL (channel));

        html = g_strdup_printf (
            "<td width=36 valign=top><a href=\"updater:channel?%d\">"
            "<img height=36 width=36 src=\"%s\" border=no></a></td>"
            "<td align=left valign=top width=50%%>"
            "<font size=+1><b><a href=\"updater:channel?%d\">%s</a></b>"
            "</font><br>%s</td>",
            channel->rcchannel.id, icon_url,
            channel->rcchannel.id, channel->rcchannel.name,
            channel->rcchannel.description ? channel->rcchannel.description :
            "");

        g_free (icon_url);

        gtk_html_write (appinfo->exec_body, stream, html, strlen (html));

        g_free (html);

        if (!first_column) {
            html = "</tr>";
            gtk_html_write (appinfo->exec_body, stream, html, strlen (html));
        }

        first_column = !first_column;
    }

    if (num_displayed && !first_column) {
        html =
            "<td width=50%% align=left colspan=2>&nbsp;</td></tr>";
        gtk_html_write (appinfo->exec_body, stream, html, strlen (html));
    }

    if (num_displayed == 0) {
        /* No channels were displayed, because we are already subscribed or
         * unsubscribed to all of them.  Give the user a nice message.
         */

        if (!subscribed) {
            if (appinfo->channel_list) {
                html = g_strdup_printf (
                    "<tr><td cellpadding=8>"
                    "%s<br>" /* You are currently subscribed to all available... */
                    "<ul>"
                    "<li>%s" /* view my sub channels */
                    "<li>%s" /* view the news page */
                    "<li>%s" /* return to the summary */
                    "</ul></td></tr>",
                    U_("You are currently subscribed to all available "
                       "channels!"),
                    U_("View my "
                       "<a href=\"updater:subscribed_page\">subscribed "
                       "channels</a>."),
                    U_("View the <a href=\"updater:news_page\">News "
                       "Page</a>."),
                    U_("Return to the "
                       "<a href=\"updater:summary\">Summary</a>."));
            }
            else {
                html = g_strdup_printf (
                    "<tr><td cellpadding=8>"
                    "%s<br>"
                    "<ul>"
                    "<li>%s" /* view the news */
                    "<li>%s" /* return to summary */
                    "</ul></td></tr>",
                    U_("There are currently no channels available to you."),
                    U_("View the <a href=\"updater:news_page\">News "
                       "Page</a>."),
                    U_("Return to the "
                       "<a href=\"updater:summary\">Summary</a>."));
            }
        } else {
            if (appinfo->channel_list) {
                html = g_strdup_printf (
                    "<tr><td cellpadding=8>"
                    "%s"
                    "<br>"
                    "<ul>"
                    "<li>%s"
                    "<li>%s"
                    "<li>%s"
                    "</ul></td></tr>",
                    U_("You are currently not subscribed to any channels!<p>"
                       "We suggest that you subscribe to the "
                       "<a href=\"updater:channel_name?Ximian "
                       "Desktop#updater:info_page\">"
                       "Ximian Desktop channel</a> to keep your Ximian "
                       "GNOME desktop running smoothly.  You can also view "
                       "the software channels available from Ximian and our "
                       "Red Carpet partners on the "
                       "<a href=\"updater:unsubscribed_page\">unsubscribed "
                       "channels page</a>."),
                    U_("View <a href=\"updater:unsubscribed_page\">"
                       "unsubscribed channels</a>."),
                    U_("View the <a href=\"updater:news_page\">News "
                       "Page</a>."),
                    U_("Return to the <a href=\"updater:summary\">"
                       "introduction to Red Carpet</a>."));
            }
            else {
                html = g_strdup_printf (
                    "<tr><td cellpadding=8>"
                    "%s"
                    "<br>"
                    "<ul>"
                    "<li>%s" /* view the news */
                    "<li>%s" /* return to the summary */
                    "</ul></td></tr>",
                    U_("There are currently no channels available to you."),
                    U_("View the <a href=\"updater:news_page\">News "
                       "Page</a>."),
                    U_("Return to the "
                       "<a href=\"updater:summary\">Summary</a>."));
            }
        }
        gtk_html_write (appinfo->exec_body, stream, html, strlen (html));
        g_free (html);
    }

    html = "</table></body></html>";
    gtk_html_write (appinfo->exec_body, stream, html, strlen (html));
    gtk_html_end (appinfo->exec_body, stream, GTK_HTML_STREAM_OK);

    rc_gui_set_page ("minor_notebook", PAGE_SUMMARY);
} /* rc_gui_show_channel_page */

void
rc_gui_show_executive_summary(void)
{
    GSList *subbed = NULL;
    GSList *i;
    GtkHTMLStream *stream;
    char *h;
    int tot_updates = 0;
    int tot_size = 0;
    int urgent = 0;
    char *size_str;
    char *n;
    RCPackageSList *necessary_updates = NULL;
    int update_desc_count;

    /* We're not dealing with individual channels right now, so lets turn off
       the subscribe menu items */
    rc_gui_html_sensitize_subscribe(FALSE);
    rc_gui_html_sensitize_unsubscribe(FALSE);

    /* Sensitize the Search menu item */
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);

    gtk_widget_set_sensitive(GET_WIDGET("verify_installed_packages"), TRUE);

    /* Hide the package info page. It can't actually show up in the
       executive summary page, but it seems to me that if we go back to the
       executive summary, when we go back to a channel, that shouldn't be
       displayed. I could probably be swayed on this, though. */
    gtk_widget_hide(GET_WIDGET("package_vbox"));

    /* Show the refresh menu item and hide the channel refresh item. The
       connection for the summary_refresh signal is in fetch_channel_list */
    rc_gui_set_refresh_item(REFRESH_SUMMARY);
    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), TRUE);

    i = appinfo->channel_list;
    while (i) {
        GUIChannel *channel = i->data;

        if (channel->subscribed)
            subbed = g_slist_append(subbed, channel);

        i = i->next;
    }

    stream = gtk_html_begin(appinfo->exec_body);

    if (!g_slist_length(subbed)) {
        h = "<html><body bgcolor=#ffffff text=#000000 marginwidth=8"
            "            marginheight=4>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        h = (char *)
            U_(
              "<font size=6><b>Welcome to Red Carpet!</b></font>"
              "<p>Thank you for using Red Carpet. From resolving "
              "package dependencies to ensuring cryptographic security, Red "
              "Carpet takes the headache out of managing your GNOME desktop "
              "system. Red Carpet is also an excellent way to keep on top of "
              "the latest software offerings. The software available to you is "
              "organized into channels; you choose the channels you want and "
              "we'll make sure they're full of great programs.</p>"
              "<p><font size=4><b>Subscribing to Channels</b></font></p>"
              "<p>Since you've just started running Red Carpet, you aren't yet "
              "subscribed to any channels. On the left of the Red Carpet window "
              "is the <b>Navigation Bar</b>, which allows you to access the "
              "main pages of Red Carpet.</p>"
              "<p>To browse available channels which you are not subscribed to, "
              "click on the <b>Unsubscribed Channels</b> icon in the navigation "
              "bar. After you've done that, to subscribe to a channel, click on "
              "it and click on the <b>Subscribe</b> button in the bottom right "
              "corner of the main content area or select <b>Subscribe</b> from "
              "the Subscriptions menu.</p>"

              "<p>We welcome your feedback and bug reports, either through "
              "the <a href=\"http://bugzilla.ximian.com/\">Ximian "
              "Bugzilla</a>, or the Red Carpet mailing list at "
              "<a href=\"http://lists.ximian.com/mailman/listinfo/red-carpet\">red-carpet@ximian.com</a>.  Thank you!</p>");
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        h = "<center><table width=100%><tr><td align=left valign=top><ul>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        h = "Take me to the "
            "<a href=\"updater:channel_name?Ximian Desktop#updater:info_page\">"
            "Ximian Desktop</a> channel now.";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        h = "</ul></td>"
            "<td align=right width=150>"
            "<img src=\"file://"SHAREDIR"/welcome.png\"></td>"
            "</tr></table></center>"
            "</body></html>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        gtk_html_end(appinfo->exec_body, stream, GTK_HTML_STREAM_OK);
        /* Special case call to make it show no buttons at all. */
        rc_gui_set_action_html(NULL, ACTION_SUMMARY, 1);
        rc_gui_set_page("minor_notebook", PAGE_SUMMARY);
        return;
    }

    /* Count the number of updates and size */
    for (i = subbed; i; i = i->next) {
        GUIChannel *channel = i->data;
        GSList *j;

        rc_debug(
            RC_DEBUG_LEVEL_INFO, "Channel \"%s\" has %d updates.\n", 
            RCCHANNEL(channel)->name,
            g_slist_length(channel->available_updates));
        tot_updates += g_slist_length(channel->available_updates);

        for (j = channel->available_updates; j; j = j->next) {
            RCPackage *package = j->data;
            RCPackageUpdate *update = package->history->data;

            rc_debug(
                RC_DEBUG_LEVEL_INFO, "   Update: %s  Severity: %d\n", 
                update->spec.name, update->importance);

            tot_size += update->package_size;

            if (update->importance == RC_IMPORTANCE_NECESSARY) {
                necessary_updates = g_slist_prepend(
                    necessary_updates, package);
            }

#if 0
            if (update->importance == RC_IMPORTANCE_NECESSARY ||
                update->importance == RC_IMPORTANCE_URGENT)
                urgent++;
#endif
        }
    }

    rc_gui_set_action_html(NULL, ACTION_SUMMARY, tot_updates ? 0 : 1);

    /* This whole block is basically a copy of most of the stuff below it,
     * just dealing with just one list.  This needs to be refactored.
     */
    if (necessary_updates) {
        GSList *pkgs;
        gchar *tmp;
        gboolean discolor = FALSE;

        /* This is the wrong place to set this */
        if (appinfo->necessary)
            g_slist_free (appinfo->necessary);
        appinfo->necessary = necessary_updates;

        gtk_widget_hide (GTK_WIDGET ("shortcut_thingie"));

        /* There are necessary updates. We need to do special display handling for these */
        tot_updates = g_slist_length (necessary_updates);
        n = rc_gui_number_to_word(tot_updates);
        h = "<html><body bgcolor=\"#ffffff\" text=\"#000000\""
            "            link=\"#000000\""
            "            marginwidth=0"
            "            marginheight=0>"
            "<table width=\"100%\" cellpadding=1 cellspacing=0 border=0>"
            "<tr>"
            "<td>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        if (tot_updates == 1) {
            h = g_strdup_printf (
                U_("There is one <b>Essential Update</b> available for your system.  To "
                   "ensure that Ximian Red Carpet continues to function properly, you must "
                   "install this update before continuing.  After installing this update, "
                   "Ximian Red Carpet will automatically restart."));
            tmp = g_strdup (U_("1 update"));
        } else {
            h = g_strdup_printf (
                U_("There are %s <b>Essential Updates</b> available for your system.  To "
                  "ensure that Ximian Red Carpet continues to function properly, you must "
                  "install this update before continuing.  After installing this update, "
                  "Ximian Red Carpet will automatically restart."), n);
            tmp = g_strdup_printf (U_("%d updates"), tot_updates);
        }

        gtk_html_write (appinfo->exec_body, stream, h, strlen (h));
        g_free (h);
        g_free (n);

        h = "<br><br></td></tr></table>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        h = rc_gui_render_yellow_bar (
            "#fcdc7b", "#000000", LEFT_CORNER_IMG, NULL,
            U_("Essential Updates"), tmp, RIGHT_CORNER_IMG);
        gtk_html_write (appinfo->exec_body, stream, h, strlen (h));
        g_free (h);
        g_free (tmp);

        h = g_strdup_printf (
            "<table width=\"100%%\" cellspacing=0 cellpadding=1>"
            "<tr bgcolor=\"#e3e5e5\">"
            "<td>&nbsp;</td>"
            "<td align=left>%s</td>" /* "Package" */
            "<td>%s</td>" /* "Version" */
            "<td>%s</td></tr>",  /*"Size"*/
            U_("Package"),
            U_("Version"),
            U_("Size"));

        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

        pkgs = g_slist_copy (necessary_updates);
        if (gui_prefs_get_pretty_names())
            pkgs = rc_package_slist_sort_by_pretty_name (pkgs);
        else
            pkgs = rc_package_slist_sort_by_name(pkgs);

        while (pkgs) {
            RCPackage *pkg = (RCPackage *) pkgs->data;
            RCPackageUpdate *update = (RCPackageUpdate *) 
                pkg->history->data;
            RCPackageUpdateSList *iter;
            RCPackage *sys_pkg = g_new0 (RCPackage, 1);
            gchar *sz_str;
            char *version;
            gchar *icon_url;

            sz_str = rc_gui_format_size_with_nbsp(
                update->package_size);
            version = rc_package_spec_version_to_str(
                RC_PACKAGE_SPEC(pkg));
            
            icon_url = rc_gui_get_mirrored_url(
                pkg->channel->icon_file,
                RCCHANNEL(pkg->channel));
            h = g_strdup_printf(
                "<tr %s valign=top>"
                // ch icon
                "<td width=18>"
                "<img src=\"%s\" width=16 height=16 border=no>"
                "</td>"
                // name
                "<td width=\"45%%\">%s</td>"
                // version
                "<td width=\"40%%\">%s</td>"
                // size
                "<td width=\"15%%\">%s</td>",
                discolor ? "bgcolor=\"#f0f2f2\"" : "",
                icon_url,
                rc_gui_get_pretty_name (pkg->spec.name),
                version,
                sz_str);
            gtk_html_write(
                appinfo->update_html, stream, h, strlen(h));
            g_free (h);
            g_free (sz_str);
            g_free(version);

            h = g_strdup_printf(
                "<tr %s><td>&nbsp;</td>"
                "<td colspan=4 valign=\"top\">",
                discolor ? "bgcolor=\"#f0f2f2\"" : "");
            gtk_html_write(
                appinfo->update_html, stream, h, strlen(h));
            g_free(h);

            sys_pkg->spec.name = g_strdup (pkg->spec.name);
            sys_pkg = rc_packman_query (appinfo->pman, sys_pkg);
            
            update_desc_count = 0;
            iter = pkg->history;
            while (iter) {
                /* Iterate through the updates in a package in an
                   importance level.  Let's only count the five most recent,
                   not including urgent or necessary updates. */
                RCPackageUpdate *this_update = iter->data;
                
                if (rc_packman_version_compare(
                        appinfo->pman,
                        RC_PACKAGE_SPEC(this_update), 
                        RC_PACKAGE_SPEC(sys_pkg)) > 0 &&
                    update_desc_count < 5) {

                    h = g_strdup_printf(
                        "<li>%s", this_update->description);
                    gtk_html_write(
                        appinfo->update_html, stream, h, 
                        strlen(h));
                    g_free(h);
                    /* Lower importances are numerically greater.
                       This makes no sense. */
                    if (this_update->importance > RC_IMPORTANCE_URGENT)
                        update_desc_count++;
                }
                iter = iter->next;
            }

            if (update_desc_count == 5) {
                h = "<li>(older updates not shown)";
                gtk_html_write(
                    appinfo->update_html, stream, h, strlen(h));
            }
            
            rc_package_free (sys_pkg);
            
            h = "</td></tr>";
            gtk_html_write(
                appinfo->update_html, stream, h, strlen(h));

            discolor = !discolor;
            pkgs = pkgs->next;
        }
        g_slist_free (pkgs);

        /* end */
        h = "</table></body></html>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        gtk_html_end(appinfo->exec_body, stream, GTK_HTML_STREAM_OK);
        rc_gui_set_page ("minor_notebook", PAGE_SUMMARY);

        return;
    }

    /* Show the navbar, in case it was hidden by the disabled free service */
    gtk_widget_show (GET_WIDGET ("shortcut_thingie"));

    size_str = rc_gui_format_size(tot_size);
    n = rc_gui_number_to_word(tot_updates);
    h = "<html><body bgcolor=\"#ffffff\" text=\"#000000\""
        "            link=\"#000000\""
        "            marginwidth=0"
        "            marginheight=0>"
        "<table width=\"100%\" cellpadding=1 cellspacing=0 border=0>"
        "<tr>"
        "<td>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

    if (tot_updates == 1) {
        h = g_strdup_printf (
            U_("There is <b>one update</b> available for your "
               "system, totalling <b>%s</b> of data to be downloaded."),
            size_str);
    } else {
        h = g_strdup_printf (
            U_("There are <b>%s updates</b> available for your "
               "system, totalling <b>%s</b> of data to be downloaded."),
            n, size_str);
    }
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    g_free (h);
    g_free(size_str);
    g_free(n);

    if (appinfo->not_root) {
        h = (char *)
            U_("<!--html-->"
               "  To update your system, you will have to run Red Carpet as root.");
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    }

#if 0
    if (urgent) {
        n = rc_gui_number_to_word(urgent);
        if (urgent == 1) {
            h = (char *) U_(" Of these updates, <b>one</b> is <b>urgent</b>.");
            gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
        } else {
            h = g_strdup_printf (
                U_(" Of these updates, <b>%s</b> are <b>urgent</b>."),
                n);
            gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
            g_free (h);
        }
        g_free(n);
    }
    h = "<br><br></td></tr></table>";
    gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
#endif

    {
        GSList *channel_update_lists = NULL;
        GSList *channel_icon_urls = NULL;
        GSList *chiter, *upiter, *iciter;
        GSList *iter;
        int cur_imp;
        int has_updates[RC_IMPORTANCE_LAST] = { 0 };

        chiter = subbed;
        while (chiter) {
            GUIChannel *ch = (GUIChannel *) chiter->data;
            RCPackageSList **imps;
            int total_size;
            int j;

            rc_gui_split_by_package_importance(
                ch->available_updates, ch->installed_packages, &imps,
                &total_size);
            channel_update_lists = g_slist_append (channel_update_lists, imps);
            channel_icon_urls = g_slist_append(
                channel_icon_urls,
                rc_gui_get_mirrored_url(
                    RCCHANNEL(ch)->icon_file, RCCHANNEL(ch)));

            for (j = 0; j < RC_IMPORTANCE_LAST; j++) {
                if (imps[j]) {
                    int p = g_slist_length(imps[j]);

                    has_updates[j] += p;

                    if (j == RC_IMPORTANCE_NECESSARY ||
                        j == RC_IMPORTANCE_URGENT)
                        urgent += p;
                }
            }
            chiter = chiter->next;
        }

        if (urgent) {
            n = rc_gui_number_to_word(urgent);
            if (urgent == 1) {
                h = (char *) U_(" Of these updates, <b>one</b> is <b>urgent</b>.");
                gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
            } else {
                h = g_strdup_printf (
                    U_(" Of these updates, <b>%s</b> are <b>urgent</b>."),
                    n);
                gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
                g_free (h);
            }
            g_free(n);
        }
        h = "<br><br></td></tr></table>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));


        for (cur_imp = 0; cur_imp < RC_IMPORTANCE_LAST; cur_imp++) {
            const char *importance_names[RC_IMPORTANCE_LAST] = {
                U_("Essential Updates"),
                U_("Urgent Updates"),
                U_("Suggested Updates"),
                U_("Feature Enhancements"),
                U_("Minor Fixes/Updates")
            };

            if (has_updates[cur_imp]) {
                GSList *deferred_lines = NULL;
                gchar *tmp;
                gboolean discolor;

                if (has_updates[cur_imp] == 1) {
                    tmp = g_strdup (U_("1 update"));
                } else {
                    tmp = g_strdup_printf (U_("%d updates"),
                                           has_updates[cur_imp]);
                }

                h = rc_gui_render_yellow_bar (
                    "#fcdc7b", "#000000", LEFT_CORNER_IMG, NULL,
                    importance_names[cur_imp], tmp, RIGHT_CORNER_IMG);
                gtk_html_write (
                    appinfo->exec_body, stream, h, strlen (h));
                g_free (h);
                g_free (tmp);

                h = g_strdup_printf (
                    "<table width=\"100%%\" cellspacing=0 cellpadding=1>"
                    "<tr bgcolor=\"#e3e5e5\">"
                    "<td>&nbsp;</td>"
                    "<td align=left>%s</td>" /* "Package" */
                    "<td>%s</td>" /* "Version" */
                    "<td>%s</td></tr>",  /*"Size"*/
                    U_("Package"),
                    U_("Version"),
                    U_("Size"));

                gtk_html_write(appinfo->exec_body, stream, h, strlen(h));

                discolor = FALSE;
                /* Now fill out the list of packages */
                for (chiter = subbed, upiter = channel_update_lists, 
                         iciter = channel_icon_urls;
                     chiter;
                     chiter = chiter->next, upiter = upiter->next, 
                         iciter = iciter->next)
                {
                    GUIChannel *ch = (GUIChannel *) chiter->data;
                    RCPackageSList **imps = (RCPackageSList **) upiter->data;
                    gchar *icon_url = (gchar *) iciter->data;

                    /* XXX this should be configurable */
                    int num_max = 4;

                    GSList *pkgs;

                    pkgs = imps[cur_imp];
                    if (gui_prefs_get_pretty_names())
                        pkgs = rc_package_slist_sort_by_pretty_name (pkgs);
                    else
                        pkgs = rc_package_slist_sort_by_name(pkgs);

                    while (pkgs) {
                        RCPackage *pkg = (RCPackage *) pkgs->data;
                        RCPackageUpdate *update = (RCPackageUpdate *) 
                            pkg->history->data;
                        RCPackage *sys_pkg = g_new0 (RCPackage, 1);
                        gchar *sz_str;
                        char *version;

                        sz_str = rc_gui_format_size_with_nbsp(
                            update->package_size);
                        version = rc_package_spec_version_to_str(
                            RC_PACKAGE_SPEC(pkg));

                        h = g_strdup_printf(
                            "<tr %s valign=top>"
                            // ch icon
                            "<td width=18><a href=\"updater:channel?%d\">"
                            "<img src=\"%s\" width=16 height=16 border=no>"
                            "</a></td>"
                            // name
                            "<td width=\"45%%\">%s</td>"
                            // version
                            "<td width=\"40%%\">%s</td>"
                            // size
                            "<td width=\"15%%\">%s</td></tr>",

                            discolor ? "bgcolor=\"#f0f2f2\"" : "",

                            RCCHANNEL(ch)->id, icon_url,

                            rc_gui_get_pretty_name (pkg->spec.name),
                            version,
                            sz_str);
                        gtk_html_write(
                            appinfo->update_html, stream, h, strlen(h));
                        g_free (h);
                        g_free (sz_str);
                        g_free(version);

                        h = g_strdup_printf(
                            "<tr %s><td>&nbsp;</td>"
                            "<td colspan=4 valign=\"top\">",
                            discolor ? "bgcolor=\"#f0f2f2\"" : "");
                        gtk_html_write(
                            appinfo->update_html, stream, h, strlen(h));
                        g_free(h);

                        sys_pkg->spec.name = g_strdup (pkg->spec.name);
                        sys_pkg = rc_packman_query (appinfo->pman, sys_pkg);

                        update_desc_count = 0;
                        iter = pkg->history;
                        while (iter) {
                            /* Iterate through the updates in a package in an
                               importance level.  Let's only count the five
                               most recent, not including urgent or necessary
                               updates. */
                            RCPackageUpdate *this_update = iter->data;

                            if (rc_packman_version_compare(
                                    appinfo->pman,
                                    RC_PACKAGE_SPEC(this_update), 
                                    RC_PACKAGE_SPEC(sys_pkg)) > 0 &&
                                update_desc_count < 5) {

                                h = g_strdup_printf(
                                    "<li>%s", this_update->description);
                                gtk_html_write(
                                    appinfo->update_html, stream, h, 
                                    strlen(h));
                                g_free(h);
                                /* Lower importances are numerically greater.
                                   This makes no sense. */
                                if (this_update->importance > 
                                    RC_IMPORTANCE_URGENT)
                                    update_desc_count++;
                            }
                            iter = iter->next;
                        }

                        if (update_desc_count == 5) {
                            h = "<li>(older updates not shown)";
                            gtk_html_write(
                                appinfo->update_html, stream, h, strlen(h));
                        }

                        rc_package_free (sys_pkg);

                        h = "</td></tr>";
                        gtk_html_write(
                            appinfo->update_html, stream, h, strlen(h));

                        discolor = !discolor;
                        pkgs = pkgs->next;

                        if (cur_imp > 1 && num_max != -1 && !--num_max) {
                            /* We need to defer the rest of these. */
                            int left_packages;
                            left_packages = g_slist_length (pkgs);
                            if (left_packages < 3) {
                                /* Bah. just keep printing the other two here.
                                   4 .. 6 will get printed. */
                                /* This will mean that the line will always 
                                   read 3 or more packages.. */
                                num_max = -1;
                                continue;
                            }

                            if (left_packages != 0) {
                                gchar *h1, *h2;
                                h1 = g_strdup_printf (
                                    "<td width=18>"
                                    "<a href=\"updater:channel?%d\">"
                                    "<img src=\"%s\" width=16 height=16 "
                                    "     border=no></a></td>"
                                    "<td colspan=4>"
                                    "<a href=\"updater:channel?%d\">",
                                    RCCHANNEL(ch)->id, icon_url,
                                    RCCHANNEL(ch)->id);

                                if (left_packages == 1) {
                                    h2 = g_strdup_printf (
                                        U_("1 other in the %s channel..."),
                                        RCCHANNEL(ch)->name);
                                } else {
                                    h2 = g_strdup_printf (
                                        U_("%d others in the %s channel..."),
                                        left_packages,
                                        RCCHANNEL(ch)->name);
                                }
                                
                                h = g_strconcat (h1, h2, "</a></td>", NULL);
                                g_free (h1);
                                g_free (h2);

                                deferred_lines = g_slist_append(
                                    deferred_lines, h);
                            }
                            break;
                        }
                    }
                }

                if (deferred_lines) {
                    GSList *defer_iter = deferred_lines;
                    while (defer_iter) {
                        char *h1;

                        h1 = g_strdup_printf(
                            "<tr %s>%s</tr>",
                            discolor ? "bgcolor=\"#f0f2f2\"" : "",
                            (char *) defer_iter->data);

                        gtk_html_write(
                            appinfo->update_html, stream, h1, strlen(h1));
                        g_free (defer_iter->data);
                        g_free(h1);
                        defer_iter = defer_iter->next;
                    }
                    g_slist_free (deferred_lines);
                }

                h = "</table><br>";
                gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
            } /* if has_updates */
        } /* importance # loop */

        /* cleanup */
        for (upiter = channel_update_lists, iciter = channel_icon_urls;
             upiter;
             upiter = upiter->next, iciter = iciter->next)
        {
            RCPackageSList **imps = (RCPackageSList **) upiter->data;
            gchar *icon_url = (gchar *) iciter->data;
            int j;
            for (j = 0; j < RC_IMPORTANCE_LAST; j++)
                g_slist_free (imps[j]);
            g_free (imps);
            g_free (icon_url);
        }

        g_slist_free (channel_update_lists);
        g_slist_free (channel_icon_urls);
        g_slist_free (subbed);
        
        /* end */
        h = "</table></body></html>";
        gtk_html_write(appinfo->exec_body, stream, h, strlen(h));
    }

    gtk_html_end(appinfo->exec_body, stream, GTK_HTML_STREAM_OK);

    rc_gui_set_page ("minor_notebook", PAGE_SUMMARY);
} /* rc_gui_show_executive_summary */

/* CB */
gboolean
rc_gui_html_object_requested(GtkHTML *html, GtkHTMLEmbedded *eb)
{
    GtkWidget *w = NULL;

    if (!strncmp (eb->classid, "rcbutton:", strlen ("rcbutton:"))) {
        gchar *bptr = eb->classid + strlen ("rcbutton:");
        GUIHtmlButtonCallbackStruct *bcs;

        bcs = (GUIHtmlButtonCallbackStruct *) strtoul (bptr, NULL, 16);

        w = gtk_button_new_with_label (bcs->button_title);
        gtk_signal_connect(
            GTK_OBJECT (w), "clicked", rc_gui_handle_uri_button_cb, bcs);
        gtk_widget_set_sensitive (w, bcs->init_sensitive);
        gtk_widget_show (w);

        gtk_object_set_data(GTK_OBJECT(html), "action_button", (gpointer) w);
    } else if (!strncmp (eb->classid, "updater:", strlen ("updater:"))) {
        rc_gui_handle_uri(html, eb->classid, NULL);
    } else {
        w = gtk_object_get_data(GTK_OBJECT(html), eb->classid);
    }

    if (!w) {
        return FALSE;
    }

    gtk_container_add(GTK_CONTAINER(eb), w);
    gtk_widget_show(GTK_WIDGET(eb));

    return TRUE;
} /* object_requested */

void
rc_gui_html_sensitize_subscribe(gboolean on)
{
    GtkWidget *w;
    
    /* Menu item */
    w = GET_WIDGET("subscribe1");
    gtk_widget_set_sensitive(w, on);
} /* rc_gui_html_sensitize_subscribe */

void
rc_gui_html_sensitize_unsubscribe(gboolean on)
{
    GtkWidget *w;
    
    /* Menu item */
    w = GET_WIDGET("unsubscribe1");
    gtk_widget_set_sensitive(w, on);
} /* rc_gui_html_sensitize_unsubscribe */

static GUIHtmlButtonCallbackStruct *
rc_gui_make_button_callback_struct (GtkHTML *html,
                                    const char *button_title,
                                    const char *uri,
                                    GUIChannel *channel,
                                    gboolean init_sensitive)
{
    GUIHtmlButtonCallbackStruct *bcs;

    bcs = g_new (GUIHtmlButtonCallbackStruct, 1);
    bcs->html = html;
    bcs->button_title = button_title;
    bcs->uri = uri;
    bcs->channel = channel;
    bcs->init_sensitive = init_sensitive;

    return bcs;
}

static void
rc_gui_handle_uri_button_cb (GtkWidget *button,
                             gpointer data)
{
    GUIHtmlButtonCallbackStruct *bcs = (GUIHtmlButtonCallbackStruct *) data;

    rc_gui_handle_uri (bcs->html, bcs->uri, bcs->channel);
}
